import dataclasses
import itertools
import json
import os
from collections import defaultdict, Counter
from pathlib import Path
from typing import cast, Any
from collections.abc import Iterable

import hydra
from hydra.utils import instantiate

from kazu.data.data import Entity, Document, IdsAndSource
from kazu.pipeline import Pipeline
from kazu.utils.constants import HYDRA_VERSION_BASE
from kazu.utils.grouping import sort_then_group


class AcceptanceTestFailure(Exception):
    pass


AcceptanceCriteria = dict[str, dict[str, dict[str, float]]]


def acceptance_criteria() -> AcceptanceCriteria:
    with open(Path(os.environ["KAZU_MODEL_PACK"]).joinpath("acceptance_criteria.json")) as f:
        data = cast(AcceptanceCriteria, json.load(f))
    return data


@hydra.main(version_base=HYDRA_VERSION_BASE, config_path=".", config_name="config")
def execute_full_pipeline_acceptance_test(cfg):
    manager = instantiate(cfg.LabelStudioManager)
    pipeline: Pipeline = instantiate(cfg.Pipeline)
    analyse_full_pipeline(pipeline, manager.export_from_ls(), acceptance_criteria())


class SectionScorer:
    def __init__(
        self,
        task: str,
        gold_ents: list[Entity],
        test_ents: list[Entity],
    ):
        """
        :param gold_ents:
        :param test_ents:
        """
        self.task = task
        self.gold_ents = gold_ents
        self.test_ents = test_ents
        self.gold_to_test_ent_soft: dict[Entity, set[Entity]] = defaultdict(set)

        self.ner_fp_soft: set[Entity] = set(test_ents)
        self.ner_fn_soft: set[Entity] = set(gold_ents)
        self.calculate_ner_matches()

        self.gold_to_test_mappings: dict[tuple[Entity, str], dict[str, IdsAndSource]] = defaultdict(
            dict
        )
        self.calculate_linking_matches()

    @staticmethod
    def group_mappings_by_source(ents: Iterable[Entity]) -> dict[str, IdsAndSource]:
        mappings_by_source = defaultdict(set)
        for ent in ents:
            for mapping in ent.mappings:
                mappings_by_source[mapping.source].add(
                    (
                        mapping.source,
                        f"{mapping.default_label}|{mapping.idx}",
                    )
                )
        return dict(mappings_by_source)

    def calculate_ner_matches(self):
        combos = itertools.product(self.gold_ents, self.test_ents)
        for (gold_ent, test_ent) in combos:
            if (
                gold_ent.spans == test_ent.spans or gold_ent.is_partially_overlapped(test_ent)
            ) and gold_ent.entity_class == test_ent.entity_class:
                self.gold_to_test_ent_soft[gold_ent].add(test_ent)
                self.ner_fp_soft.discard(test_ent)
                self.ner_fn_soft.discard(gold_ent)

    def calculate_linking_matches(self):
        for gold_ent, test_ents in self.gold_to_test_ent_soft.items():
            gold_mappings_by_source = self.group_mappings_by_source([gold_ent])
            test_mappings_by_source = self.group_mappings_by_source(test_ents)
            sources = set(gold_mappings_by_source.keys()).union(test_mappings_by_source.keys())
            for source in sources:
                gold_mappings = gold_mappings_by_source.get(source, set())
                test_mappings_set = test_mappings_by_source.get(source, set())
                tp = gold_mappings.intersection(test_mappings_set)
                fn = gold_mappings - test_mappings_set
                fp = test_mappings_set - gold_mappings
                self.gold_to_test_mappings[
                    (
                        gold_ent,
                        source,
                    )
                ] = {"tp": tp, "fp": fp, "fn": fn}


def score_sections(
    docs: list[Document],
) -> dict[str, list[SectionScorer]]:
    """Score a list of documents by Section.

    :param docs:
    :return: dict of entity class to one scorer per section
    """

    result = defaultdict(list)
    for doc in docs:
        for section in doc.sections:
            gold_ents: list[Entity] = section.metadata["gold_entities"]
            gold_ents_by_class = {
                k: list(v) for k, v in sort_then_group(gold_ents, key_func=lambda x: x.entity_class)
            }
            for entity_class, test_ents in sort_then_group(
                section.entities, key_func=lambda x: x.entity_class
            ):
                scorer = SectionScorer(
                    task=section.metadata["label_studio_task_id"],
                    gold_ents=gold_ents_by_class.get(entity_class, []),
                    test_ents=list(test_ents),
                )
                result[entity_class].append(scorer)

    return dict(result)


@dataclasses.dataclass
class AggregatedAccuracyResult:
    tp: int = 0
    fp: int = 0
    fn: int = 0
    fp_counter: Counter = dataclasses.field(default_factory=Counter)
    fn_counter: Counter = dataclasses.field(default_factory=Counter)
    fp_items_to_tasks: dict[Any, set[str]] = dataclasses.field(default_factory=dict)
    fn_items_to_tasks: dict[Any, set[str]] = dataclasses.field(default_factory=dict)

    def add_fp(self, item: Any, task: str) -> None:
        self.fp += 1
        self.fp_counter[item] += 1
        self.fp_items_to_tasks.setdefault(item, set()).add(task)

    def add_fn(self, item: Any, task: str) -> None:
        self.fn += 1
        self.fn_counter[item] += 1
        self.fn_items_to_tasks.setdefault(item, set()).add(task)

    def tasks_for_fp(self, items: list[Any]) -> Iterable[str]:
        return (
            task
            for item, tasks in self.fp_items_to_tasks.items()
            for task in tasks
            if item in items
        )

    def tasks_for_fn(self, items: list[Any]) -> Iterable[str]:
        return (
            task
            for item, tasks in self.fn_items_to_tasks.items()
            for task in tasks
            if item in items
        )

    @property
    def precision(self) -> float:
        div = self.fp + self.tp
        if div == 0:
            return 0.0
        else:
            return float(self.tp / div)

    @property
    def recall(self) -> float:
        div = self.fn + self.tp
        if div == 0:
            return 0.0
        else:
            return float(self.tp / div)

    @property
    def fp_info(self) -> list[Any]:
        return self.fp_counter.most_common()

    @property
    def fn_info(self) -> list[Any]:
        return self.fn_counter.most_common()


def aggregate_ner_results(
    class_and_scorers: dict[str, list[SectionScorer]]
) -> dict[str, AggregatedAccuracyResult]:
    result: dict[str, AggregatedAccuracyResult] = {}
    for ent_class, scorers in class_and_scorers.items():
        acc_result = AggregatedAccuracyResult()
        for scorer in scorers:
            acc_result.tp += len(scorer.gold_to_test_ent_soft)
            for ent in scorer.ner_fn_soft:
                acc_result.add_fn(item=ent.match, task=scorer.task)
            for ent in scorer.ner_fp_soft:
                acc_result.add_fp(item=ent.match, task=scorer.task)
        result[ent_class] = acc_result
    return result


def aggregate_linking_results(
    class_and_scorers: dict[str, list[SectionScorer]]
) -> dict[str, AggregatedAccuracyResult]:

    result: defaultdict[str, AggregatedAccuracyResult] = defaultdict(AggregatedAccuracyResult)
    for _, scorers in class_and_scorers.items():
        for scorer in scorers:
            for (
                _,
                source,
            ), mapping_result in scorer.gold_to_test_mappings.items():
                acc_result = result[source]
                acc_result.tp += len(mapping_result["tp"])
                for fn_item in mapping_result["fn"]:
                    acc_result.add_fn(item=fn_item, task=scorer.task)
                for fp_item in mapping_result["fp"]:
                    acc_result.add_fp(item=fp_item, task=scorer.task)
    return dict(result)


def check_results_meet_threshold(
    results: dict[str, AggregatedAccuracyResult],
    thresholds: dict[str, dict[str, float]],
) -> None:

    for key, threshold in thresholds.items():
        aggregated_result = results[key]
        prec = aggregated_result.precision
        rec = aggregated_result.recall
        message = "\n".join(f"{k} <incorrect {v} times>" for k, v in aggregated_result.fp_info)
        prec_message = (
            f"{aggregated_result.tp} / {aggregated_result.fp + aggregated_result.tp} \n{message}"
        )
        if prec >= threshold["precision"]:
            print(f"{key} passed precision threshold: {prec}. {prec_message}\n\n")
        else:
            raise AcceptanceTestFailure(
                f"{key} failed to meet precision threshold: {prec}. {prec_message}"
            )

        message = "\n".join(f"{k} <missed {v} times>" for k, v in aggregated_result.fn_info)
        rec_message = (
            f"{aggregated_result.tp} / {aggregated_result.fn + aggregated_result.tp} \n{message}"
        )
        if rec >= threshold["recall"]:
            print(f"{key} passed recall threshold: {rec}. {rec_message}\n\n")
        else:
            raise AcceptanceTestFailure(
                f"{key} failed to meet recall threshold: {rec}. {rec_message}"
            )


def analyse_full_pipeline(
    pipeline: Pipeline,
    docs: list[Document],
    acceptance_criteria: dict[str, dict[str, dict[str, float]]],
) -> None:
    pipeline(docs)
    ner_dict = score_sections(docs)
    ner_results = aggregate_ner_results(ner_dict)
    check_results_meet_threshold(
        results=ner_results, thresholds=acceptance_criteria["NER_THRESHOLDS"]
    )

    linking_results = aggregate_linking_results(ner_dict)
    check_results_meet_threshold(
        results=linking_results, thresholds=acceptance_criteria["LINKING_THRESHOLDS"]
    )


@hydra.main(version_base=HYDRA_VERSION_BASE, config_path="../../", config_name="conf")
def check_annotation_consistency(cfg):

    manager = instantiate(cfg.LabelStudioManager)
    docs = manager.export_from_ls()
    all_ents: list[Entity] = []
    ent_to_task_lookup: dict[Entity, str] = {}  # used for reporting task id that may have issues
    for doc in docs:
        for section in doc.sections:
            ents: list[Entity] = section.metadata["gold_entities"]
            all_ents.extend(ents)
            ent_to_task_lookup.update(
                {ent: str(section.metadata["label_studio_task_id"]) for ent in ents}
            )

    messages: defaultdict[str, set[str]] = defaultdict(set)
    # update the messages dict with any apparent issues
    for match_str, ents_iter in sort_then_group(all_ents, lambda x: x.match):
        ents = list(ents_iter)
        check_ent_match_abnormalities(ent_to_task_lookup, ents, match_str, messages)
        check_ent_class_consistency(ent_to_task_lookup, ents, match_str, messages)
        check_ent_mapping_consistency(ent_to_task_lookup, ents, match_str, messages)

    print(f"{len(messages)} tasks with issues:\n")
    for doc_id in sorted(messages):
        print(f"\ntask id: {doc_id}\n" + "*" * 20 + "\n" + "\n".join(messages[doc_id]))


def check_ent_match_abnormalities(
    ent_to_task_lookup: dict[Entity, str],
    ents: list[Entity],
    match_str: str,
    messages: dict[str, set[str]],
) -> None:
    """Checks to see if any gold standard spans look a bit weird.

    :param ent_to_task_lookup:
    :param ents:
    :param match_str:
    :param messages:
    :return:
    """
    if len(match_str) == 1 or (
        len(match_str) == 2 and not all(char.isalnum() for char in match_str)
    ):
        for ent in ents:
            messages[ent_to_task_lookup[ent]].add(
                f"WARNING: ent string <{match_str}> may be abnormal"
            )


def check_ent_class_consistency(
    ent_to_task_lookup: dict[Entity, str],
    ents: list[Entity],
    match_str: str,
    messages: dict[str, set[str]],
) -> None:
    """Checks to see if any match strings have different entity_class
    information.

    :param ent_to_task_lookup:
    :param ents:
    :param match_str:
    :param messages:
    :return:
    """
    by_doc = defaultdict(set)
    all_entity_classes = set()
    for ent in ents:
        by_doc[ent_to_task_lookup[ent]].add(ent.entity_class)
        all_entity_classes.add(ent.entity_class)

    if len(all_entity_classes) > 1:
        message = "\n".join(f"{doc_id}:{classes}" for doc_id, classes in by_doc.items())
        for doc_id, classes in by_doc.items():
            messages[doc_id].add(
                f"WARNING: ent string <{match_str}> is class confused in the following documents:\n{message}"
            )


def check_ent_mapping_consistency(
    ent_to_task_lookup: dict[Entity, str],
    ents: list[Entity],
    match_str: str,
    messages: dict[str, set[str]],
) -> None:
    """Checks to see if any entity string matches have inconsistent mapping
    information.

    :param ent_to_task_lookup:
    :param ents:
    :param match_str:
    :param messages:
    :return:
    """
    mappings_to_doc_id = defaultdict(set)
    for ent in ents:
        mappings_to_doc_id[frozenset(ent.mappings)].add(ent_to_task_lookup[ent])
    if len(mappings_to_doc_id) > 1:
        group_definitions = set()
        doc_id_to_groups = defaultdict(set)
        for i, (mappings, doc_ids) in enumerate(mappings_to_doc_id.items()):
            group_key = f"group_{i}:\n"
            group_msg = (
                "\n".join(
                    f"{mapping.source}|{mapping.default_label}|{mapping.idx}"
                    for mapping in mappings
                )
                if len(mappings) > 0
                else "<NONE>"
            )
            group_msg = group_key + group_msg + "\n"
            group_definitions.add(group_msg)
            for doc_id in doc_ids:
                doc_id_to_groups[doc_id].add(group_key)

        group_definition_message = "\n".join(group_definitions)
        confused_docs_message = "\n".join(
            f"{doc_id}:{sorted(groups)}" for doc_id, groups in sorted(doc_id_to_groups.items())
        )

        overall_message = (
            f"WARNING: ent string <{match_str}> has inconsistent mappings. This may be a genuine ambiguity or "
            f"a mistake in annotation: \n\n {group_definition_message} \n\naffected tasks:"
            f"\n {confused_docs_message}\n"
        )
        for doc_id in doc_id_to_groups:
            messages[doc_id].add(overall_message)


if __name__ == "__main__":
    execute_full_pipeline_acceptance_test()



import logging
from collections import defaultdict
from enum import auto
from typing import Any, Literal, Optional, Iterable

from spacy.matcher import Matcher
from spacy.tokens import Span


from kazu.data.data import Document, Entity, Section, AutoNameEnum
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.steps import document_iterating_step
from kazu.steps import ParserDependentStep
from kazu.utils.spacy_pipeline import (
    SpacyPipelines,
    BASIC_PIPELINE_NAME,
    basic_spacy_pipeline,
)
from kazu.utils.spacy_object_mapper import SpacyToKazuObjectMapper

SpacyMatcherRules = list[list[dict[str, Any]]]
logger = logging.getLogger(__name__)
TPOrFP = Literal["tp", "fp"]
TPOrFPMatcher = dict[TPOrFP, Matcher]
MatcherMentionRules = dict[str, dict[str, dict[TPOrFP, SpacyMatcherRules]]]
MatcherClassRules = dict[str, dict[TPOrFP, SpacyMatcherRules]]
MentionMatchers = dict[str, dict[str, TPOrFPMatcher]]
ClassMatchers = dict[str, TPOrFPMatcher]
_KeyToMatcherResults = dict[tuple[str, str], bool]


class MatcherResult(AutoNameEnum):
    HIT = auto()
    MISS = auto()
    NOT_CONFIGURED = auto()


class RulesBasedEntityClassDisambiguationFilterStep(ParserDependentStep):
    """Removes instances of :class:`.Entity` from
    :class:`.Section`\\s that don't meet rules based
    disambiguation requirements in at least one location in the document.

    This step utilises Spacy `Matcher <https://spacy.io/api/matcher>`_
    rules to determine whether an entity class and or/mention entities are valid or not.
    These Matcher rules operate on the sentence in which each mention under consideration
    is located.

    Rules can have both true positive and false positive aspects. If defined, that
    aspect MUST be correct at least once in the document for all entities with the same
    key (composed of the matched string and entity class) to be valid.

    Non-contiguous entities are evaluated on the full span of the text they cover, rather
    than the specific tokens.
    """

    def __init__(
        self,
        class_matcher_rules: MatcherClassRules,
        mention_matcher_rules: MatcherMentionRules,
        parsers: Iterable[OntologyParser],
    ):
        """

        :param class_matcher_rules: these should follow the format:

            .. code-block:: python

                {
                    "<entity class>": {
                        "<tp or fp (for true positive or false positive rules respectively>": [
                            "<a list of rules>",
                            "<according to the spacy pattern matcher syntax>",
                        ]
                    }
                }
        :param mention_matcher_rules: these should follow the format:

            .. code-block:: python

                {
                    "<entity class>": {
                        "<mention to disambiguate>": {
                            "<tp or fp>": [
                                "<a list of rules>",
                                "<according to the spacy pattern matcher syntax>",
                            ]
                        }
                    }
                }
        """
        super().__init__(parsers)
        self.mapper = SpacyToKazuObjectMapper(parsers)
        self.class_matcher_rules = class_matcher_rules
        self.mention_matcher_rules = mention_matcher_rules
        self.spacy_pipelines = SpacyPipelines()
        self.spacy_pipelines.add_from_func(BASIC_PIPELINE_NAME, basic_spacy_pipeline)
        self.spacy_pipelines.add_reload_callback_func(
            BASIC_PIPELINE_NAME, self._build_class_matchers
        )
        self.spacy_pipelines.add_reload_callback_func(
            BASIC_PIPELINE_NAME, self._build_mention_matchers
        )
        self._build_class_matchers()
        self._build_mention_matchers()

    def _build_class_matchers(self) -> None:
        result: ClassMatchers = {}
        for class_name, rules in self.class_matcher_rules.items():
            for rule_type, rule_instances in rules.items():
                if rule_instances is not None:
                    matcher = Matcher(self.spacy_pipelines.get_model(BASIC_PIPELINE_NAME).vocab)
                    matcher.add(f"{class_name}_{rule_type}", rule_instances)
                    result.setdefault(class_name, {})[rule_type] = matcher
        self.class_matchers = result

    def _build_mention_matchers(self) -> None:
        result: MentionMatchers = {}
        rule_type: TPOrFP
        for class_name, target_term_dict in self.mention_matcher_rules.items():
            for target_term, rules in target_term_dict.items():
                for rule_type, rule_instances in rules.items():
                    if rule_instances is not None:
                        matcher = Matcher(self.spacy_pipelines.get_model(BASIC_PIPELINE_NAME).vocab)
                        matcher.add(f"{class_name}_{target_term}_{rule_type}", rule_instances)
                        result_for_class = result.setdefault(class_name, {})
                        result_for_class_and_target_term = result_for_class.setdefault(
                            target_term, {}
                        )
                        result_for_class_and_target_term[rule_type] = matcher
        self.mention_matchers = result

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        ent_tp_class_results: _KeyToMatcherResults = {}
        ent_fp_class_results: _KeyToMatcherResults = {}
        ent_tp_mention_results: _KeyToMatcherResults = {}
        ent_fp_mention_results: _KeyToMatcherResults = {}

        class_tp_is_configured_for_key = {}
        class_fp_is_configured_for_key = {}
        mention_tp_is_configured_for_key = {}
        mention_fp_is_configured_for_key = {}

        # keep track of only entities that could be affected by this step, so we don't need
        # to loop over everything later
        section_to_ents_under_consideration: defaultdict[Section, set[Entity]] = defaultdict(set)

        for section in doc.sections:
            ent_to_span = self.mapper(section)

            for entity in section.entities:
                if entity not in ent_to_span:
                    # there's a chance that an entity can't be matched to a spacy
                    # span, e.g. if the entity only covers non-token text.
                    # We skip the entity in this case.
                    continue
                entity_class = entity.entity_class

                entity_match = entity.match
                key = (
                    entity_match,
                    entity_class,
                )
                maybe_class_matchers = self.class_matchers.get(entity_class)
                maybe_mention_matchers = self.mention_matchers.get(entity_class, {}).get(
                    entity.match
                )
                # if neither class nor mention matcher defined (the usual case), just continue
                if maybe_class_matchers is None and maybe_mention_matchers is None:
                    continue

                section_to_ents_under_consideration[section].add(entity)
                tp_class_result, fp_class_result = self._check_tp_fp_matcher_rules(
                    entity, ent_to_span, maybe_class_matchers
                )
                if tp_class_result is MatcherResult.NOT_CONFIGURED:
                    class_tp_is_configured_for_key[key] = False
                else:
                    class_tp_is_configured_for_key[key] = True
                if fp_class_result is MatcherResult.NOT_CONFIGURED:
                    class_fp_is_configured_for_key[key] = False
                else:
                    class_fp_is_configured_for_key[key] = True

                ent_tp_class_results[key] = (
                    ent_tp_class_results.get(key, False) or tp_class_result is MatcherResult.HIT
                )
                ent_fp_class_results[key] = (
                    ent_fp_class_results.get(key, False) or fp_class_result is MatcherResult.HIT
                )

                tp_mention_result, fp_mention_result = self._check_tp_fp_matcher_rules(
                    entity, ent_to_span, maybe_mention_matchers
                )
                if tp_mention_result is MatcherResult.NOT_CONFIGURED:
                    mention_tp_is_configured_for_key[key] = False
                else:
                    mention_tp_is_configured_for_key[key] = True
                if fp_mention_result is MatcherResult.NOT_CONFIGURED:
                    mention_fp_is_configured_for_key[key] = False
                else:
                    mention_fp_is_configured_for_key[key] = True

                ent_tp_mention_results[key] = (
                    ent_tp_mention_results.get(key, False) or tp_mention_result is MatcherResult.HIT
                )
                ent_fp_mention_results[key] = (
                    ent_fp_mention_results.get(key, False) or fp_class_result is MatcherResult.HIT
                )

        for section, ents in section_to_ents_under_consideration.items():
            for ent in ents:
                key = (
                    ent.match,
                    ent.entity_class,
                )
                if (
                    (class_fp_is_configured_for_key[key] and ent_fp_class_results[key])
                    or (class_tp_is_configured_for_key[key] and not ent_tp_class_results[key])
                    or (mention_fp_is_configured_for_key[key] and ent_fp_mention_results[key])
                    or (mention_tp_is_configured_for_key[key] and not ent_tp_mention_results[key])
                ):
                    section.entities.remove(ent)

    @staticmethod
    def _check_matcher(context: Span, maybe_matcher: Optional[Matcher]) -> MatcherResult:
        if maybe_matcher is None:
            return MatcherResult.NOT_CONFIGURED
        if maybe_matcher(context):
            return MatcherResult.HIT
        else:
            return MatcherResult.MISS

    @staticmethod
    def _check_tp_fp_matcher_rules(
        entity: Entity, ent_to_span: dict[Entity, Span], matchers: Optional[TPOrFPMatcher]
    ) -> tuple[MatcherResult, MatcherResult]:
        if matchers is None:
            return MatcherResult.NOT_CONFIGURED, MatcherResult.NOT_CONFIGURED
        span = ent_to_span[entity]
        context = span.sent

        tp_matcher = matchers.get("tp")
        tp_result = RulesBasedEntityClassDisambiguationFilterStep._check_matcher(
            context, tp_matcher
        )
        fp_matcher = matchers.get("fp")
        fp_result = RulesBasedEntityClassDisambiguationFilterStep._check_matcher(
            context, fp_matcher
        )
        return tp_result, fp_result



from typing import Any
from collections.abc import Iterator

from torch.utils.data import IterableDataset
from transformers import BatchEncoding


class HFDataset(IterableDataset[dict[str, Any]]):
    def __getitem__(self, index: int) -> dict[str, Any]:
        return {
            "input_ids": self.encodings.data["input_ids"][index],
            "attention_mask": self.encodings.data["attention_mask"][index],
            "token_type_ids": self.encodings.data["token_type_ids"][index],
        }

    def __init__(self, encodings: BatchEncoding):
        """Simple implementation of :class:`torch.utils.data.IterableDataset`\\
        , producing HF tokenizer input_id.

        :param encodings:
        """
        self.encodings = encodings
        self.dataset_size = len(encodings.data["input_ids"])

    def __iter__(self) -> Iterator[dict[str, Any]]:

        for i in range(self.dataset_size):
            yield {
                "input_ids": self.encodings.data["input_ids"][i],
                "attention_mask": self.encodings.data["attention_mask"][i],
                "token_type_ids": self.encodings.data["token_type_ids"][i],
            }



import logging
import hashlib
from typing import Optional
from collections.abc import Iterable

import numpy
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from kazu.utils.caching import kazu_disk_cache
from kazu.data.data import SynonymTerm, SynonymTermWithMetrics
from kazu.database.in_memory_db import (
    MetadataDatabase,
    SynonymDatabase,
    NormalisedSynonymStr,
)
from kazu.language.string_similarity_scorers import BooleanStringSimilarityScorer
from kazu.ontology_preprocessing.base import (
    OntologyParser,
)
from kazu.utils.string_normalizer import StringNormalizer
from kazu.utils.utils import create_char_ngrams

logger = logging.getLogger(__name__)


class DictionaryIndex:
    """The dictionary index looks for SynonymTerms via a char ngram search
    between the normalised version of the query string and the term_norm of all
    SynonymTerms associated with the provided OntologyParser."""

    def __init__(
        self,
        parser: OntologyParser,
        boolean_scorers: Optional[list[BooleanStringSimilarityScorer]] = None,
    ):
        """

        :param parser:
        :param boolean_scorers: precision can be increased by applying boolean checks on the returned result, for
            instance, checking that all integers are represented, that any noun modifiers are present etc
        """

        parser.populate_databases(force=False)
        self.synonym_db: SynonymDatabase = SynonymDatabase()
        self.synonyms_for_parser = self.synonym_db.get_all(parser.name)
        self.entity_class = parser.entity_class
        self.parser_name = parser.name
        self.metadata_db: MetadataDatabase = MetadataDatabase()
        self.namespace = self.__class__.__name__
        self.boolean_scorers = boolean_scorers
        self.normalized_synonyms: list[str] = []
        self.synonym_list: list[SynonymTerm] = []
        # we sort here to ensure that the order of self.synonym_list matches the
        # order of the synonyms in the cached vectorized from _build_index_cache
        # if present, without relying on the order in the SynonymDatabase being the
        # same.
        for norm_syn, syn_term in sorted(self.synonyms_for_parser.items()):
            self.normalized_synonyms.append(norm_syn)
            self.synonym_list.append(syn_term)
        self.vectorizer, self.tf_idf_matrix = self.build_index_cache()

    def apply_boolean_scorers(self, reference_term: str, query_term: str) -> bool:

        if self.boolean_scorers is not None:
            return all(
                scorer(reference_term=reference_term, query_term=query_term)
                for scorer in self.boolean_scorers
            )
        else:
            return True

    def search(self, query: str, top_n: int = 15) -> Iterable[SynonymTermWithMetrics]:
        """Search the index with a query string.

        :param query: term to search
        :param top_n: max number of results
        :return:
        """

        match_norm = StringNormalizer.normalize(query, entity_class=self.entity_class)
        exact_match_term = self.synonyms_for_parser.get(match_norm)
        if exact_match_term is not None:
            term_with_metrics = SynonymTermWithMetrics.from_synonym_term(
                exact_match_term, search_score=100.0, bool_score=True, exact_match=True
            )
            yield term_with_metrics

        else:
            # benchmarking suggests converting to dense is faster than using the
            # csr_matrix version. Strange...
            query_arr = self.vectorizer.transform([match_norm]).todense()
            # minus to negate, so arg sort works in correct order
            score_matrix = np.squeeze(-np.asarray(self.tf_idf_matrix.dot(query_arr.T)))
            neighbours = score_matrix.argsort()[:top_n]
            # don't use torch for this - it's slow
            # query = torch.FloatTensor(query)
            # score_matrix = self.tf_idf_matrix_torch.matmul(query.T)
            # score_matrix = torch.squeeze(score_matrix.T)
            # neighbours = torch.argsort(score_matrix, descending=True)[:top_n]

            distances = score_matrix[neighbours]
            distances = 100 * -distances
            for neighbour, score in zip(neighbours, distances):
                # get by index
                term = self.synonym_list[neighbour]
                if self.apply_boolean_scorers(reference_term=match_norm, query_term=term.term_norm):
                    term_with_metrics = SynonymTermWithMetrics.from_synonym_term(
                        term, search_score=score, bool_score=True, exact_match=False
                    )
                    yield term_with_metrics
                else:
                    logger.debug("filtered term %s as failed boolean checks", term)

    @kazu_disk_cache.memoize(ignore={0, 1})
    def _build_index_cache(
        self, synonyms_for_parser: Iterable[NormalisedSynonymStr], _cache_key: bytes
    ) -> tuple[TfidfVectorizer, numpy.ndarray]:
        logger.info("building TfidfVectorizer for %s", self.parser_name)
        vectorizer = TfidfVectorizer(min_df=1, analyzer=create_char_ngrams, lowercase=False)
        tf_idf_matrix = vectorizer.fit_transform(synonyms_for_parser)
        return vectorizer, tf_idf_matrix

    def build_index_cache(self) -> tuple[TfidfVectorizer, numpy.ndarray]:
        """Build the cache for the index."""

        h = hashlib.new("sha1", usedforsecurity=False)
        for norm_syn in self.normalized_synonyms:
            h.update(norm_syn.encode(encoding="utf-8"))

        return self._build_index_cache(self.normalized_synonyms, h.digest())



from typing import Any

from kazu.data.data import Document
from kazu.annotation.label_studio import (
    LabelStudioAnnotationView,
    KazuToLabelStudioConverter,
)

LSView = str
LSTask = dict[str, Any]


class LSWebUtils:
    def __init__(
        self, ls_view_generator: LabelStudioAnnotationView, ls_converter: KazuToLabelStudioConverter
    ):
        self.ls_view_generator = ls_view_generator
        self.ls_converter = ls_converter

    def kazu_doc_to_ls(self, doc: Document) -> tuple[LSView, list[LSTask]]:
        ls_tasks = list(self.ls_converter.convert_single_doc_to_tasks(doc))
        ls_view = self.ls_view_generator.create_main_view(ls_tasks)
        return ls_view, ls_tasks

    @staticmethod
    def default() -> "LSWebUtils":
        return LSWebUtils(
            LabelStudioAnnotationView.with_default_colours(), KazuToLabelStudioConverter()
        )



import logging
from collections import defaultdict
from copy import deepcopy
from functools import cached_property
from typing import Any, Optional, Union
from collections.abc import Iterable
from collections.abc import (
    Mapping as CollectionsMapping,
)  # due to name conflict with kazu.data.data.Mapping
from xml.dom.minidom import Document as XMLDocument, DOMImplementation
from xml.dom.minidom import Element, getDOMImplementation

import requests
from kazu.data.data import (
    Document,
    Section,
    Entity,
    Mapping,
    StringMatchConfidence,
    CharSpan,
)
from kazu.utils.grouping import sort_then_group
from requests import HTTPError

logger = logging.getLogger(__name__)

_TAX_NAME = "taxonomy"


class KazuToLabelStudioConverter:
    """Converts a Kazu :class:`.Document` into Label Studio tasks.

    Since LS is region based, we need to create a new region for every
    CharSpan (even overlapping ones), and add entity information (class,
    mappings etc) to the region.
    """

    @classmethod
    def convert_single_doc_to_tasks(cls, doc: Document) -> Iterable[dict]:
        doc_id = doc.idx
        for i, section in enumerate(doc.sections):
            idx = f"{doc_id}_{section.name}_{i}"
            data = {}
            data["text"] = section.text
            data["id"] = idx
            annotations = []
            result_values = cls._create_label_studio_labels(section.entities, section.text)
            annotation = {"id": idx, "result": result_values}
            annotations.append(annotation)
            yield {"data": data, "annotations": annotations}

    @classmethod
    def convert_docs_to_tasks(cls, docs: list[Document]) -> list[dict]:
        return [task for doc in docs for task in cls.convert_single_doc_to_tasks(doc)]

    @staticmethod
    def _create_label_studio_labels(
        entities: list[Entity],
        text: str,
    ) -> list[dict]:
        result_values: list[dict] = []
        for ent in entities:
            ent_hash = hash(ent)
            prev_region_id: Optional[str] = None
            if len(ent.spans) > 2:
                logger.warning(
                    """Currently we can't handle entities with 3 spans.
                    Problems would occur when we convert back to Kazu Ents for the comparison
                    (see LSToKazuConversion._create_non_contiguous_entities).
                    This is because we don't anticipate 3-span entities actually occuring, and
                    it would complicate the code to handle these.
                    Adding this warning as a safeguard"""
                )
            for span in ent.spans:
                region_id_str = f"{ent_hash}_{span}"
                match = text[span.start : span.end]
                ner_region = KazuToLabelStudioConverter._create_ner_region(
                    ent, region_id_str, span, match
                )
                result_values.append(ner_region)
                result_normalisation_value = KazuToLabelStudioConverter._create_mapping_region(
                    ent, region_id_str, span, match
                )
                result_values.append(result_normalisation_value)
                if prev_region_id is not None:
                    result_values.append(
                        KazuToLabelStudioConverter._create_non_contig_entity_links(
                            prev_region_id, region_id_str
                        )
                    )
                prev_region_id = region_id_str
        return result_values

    @staticmethod
    def _create_non_contig_entity_links(
        from_id: str, to_id: str
    ) -> dict[str, Union[str, list[str]]]:
        return {
            "from_id": from_id,
            "to_id": to_id,
            "type": "relation",
            "direction": "right",
            "labels": ["non-contig"],
        }

    @staticmethod
    def _create_mapping_region(
        ent: Entity, region_id: str, span: CharSpan, match: str
    ) -> dict[str, Any]:

        return {
            "id": region_id,
            "from_name": _TAX_NAME,
            "to_name": "text",
            "type": "taxonomy",
            "origin": "manual",
            "value": {
                "start": span.start,
                "end": span.end,
                "text": match,
                "taxonomy": sorted(
                    set(
                        (mapping.source, f"{mapping.default_label}|{mapping.idx}")
                        for mapping in ent.mappings
                    )
                )
                if len(ent.mappings) > 0
                else [("None", "unmapped|unmapped")],
            },
        }

    @staticmethod
    def _create_ner_region(
        ent: Entity, region_id: str, span: CharSpan, match: str
    ) -> dict[str, Any]:
        return {
            "id": region_id,
            "from_name": "ner",
            "to_name": "text",
            "type": "labels",
            "origin": "manual",
            "value": {
                "start": span.start,
                "end": span.end,
                "score": 1.0,
                "text": match,
                "labels": [ent.entity_class],
            },
        }


class LSToKazuConversion:
    def __init__(self, task: dict):
        self.text = task["data"]["text"]
        self.task_data_id = task["data"]["id"]
        self.label_studio_task_id = task["id"]
        if len(task["annotations"]) > 1:
            logger.warning(
                "warning: more than one annotation section. Will only use annotations from %s",
                task["annotations"][0]["id"],
            )

        annotation = task["annotations"][0]
        result = annotation["result"]

        self.id_to_charspan: dict[str, CharSpan] = {}
        self.id_to_labels: dict[str, set[str]] = defaultdict(set)
        self.id_to_mappings: dict[str, set[Mapping]] = defaultdict(set)
        self.non_contig_id_map: dict[str, set[str]] = defaultdict(set)
        self.non_contig_regions = set()

        for result_data in result:
            is_region = "id" in result_data
            if is_region:
                region_id = result_data["id"]
                data_type = result_data["type"]
                span = CharSpan(
                    start=result_data["value"]["start"], end=result_data["value"]["end"]
                )
                self.id_to_charspan[region_id] = span
                if data_type == "labels":
                    self.id_to_labels[region_id].update(result_data["value"]["labels"])
                elif data_type == "taxonomy":
                    self.id_to_mappings[region_id].update(
                        self.create_mappings(result_data["value"]["taxonomy"], region_id)
                    )
            else:
                self.non_contig_id_map[result_data["from_id"]].add(result_data["to_id"])
                self.non_contig_regions.add(result_data["from_id"])
                self.non_contig_regions.add(result_data["to_id"])

    def create_section(self) -> Section:
        section = Section(text=self.text, name=self.task_data_id)
        section.metadata["label_studio_task_id"] = self.label_studio_task_id
        section.metadata["gold_entities"] = self.create_ents()
        return section

    def create_mappings(
        self, taxonomy_hits: Iterable[tuple[str, str]], task_id: str
    ) -> list[Mapping]:
        mappings = []
        for tax in taxonomy_hits:
            if len(tax) != 2:
                logger.warning(
                    "warning! malformed taxonomy label (prob not a low level term): %s <%s>",
                    tax,
                    task_id,
                )
                continue
            source, idx_str = tax
            default_label, idx = idx_str.split("|")
            mappings.append(
                Mapping(
                    default_label=default_label,
                    source=source,
                    parser_name="gold",
                    idx=idx,
                    string_match_strategy="gold",
                    string_match_confidence=StringMatchConfidence.HIGHLY_LIKELY,
                )
            )
        return mappings

    def create_ents(self) -> list[Entity]:
        entities = []
        for region_id, span in self.id_to_charspan.items():
            if region_id not in self.non_contig_regions:
                for label in self.id_to_labels[region_id]:
                    ent = self._create_contiguous_entity(label, region_id, span)
                    entities.append(ent)
            elif region_id in self.non_contig_id_map:
                entities.extend(self._create_non_contiguous_entities(region_id))
        return entities

    def _create_non_contiguous_entities(self, region_id: str) -> Iterable[Entity]:
        for to_id in self.non_contig_id_map[region_id]:
            from_span = self.id_to_charspan[region_id]
            to_span = self.id_to_charspan[to_id]
            # we need to check from_id and to_id have matching labels
            labels = self.id_to_labels[region_id].intersection(self.id_to_labels[to_id])
            match = f"{self.text[to_span.start:to_span.end]} {self.text[from_span.start : from_span.end]}"
            spans = frozenset([to_span, from_span])
            # since independent regions might have different mapping values, we merge them all
            # ideally this wouldn't happen if human annotation is consistent
            mappings = deepcopy(self.id_to_mappings.get(region_id, set()))
            mappings.update(self.id_to_mappings.get(to_id, set()))
            for label in labels:
                yield Entity(
                    match=match,
                    entity_class=label,
                    spans=spans,
                    namespace="gold",
                    mappings=mappings,
                )

    def _create_contiguous_entity(self, label, region_id, span):
        single_span = frozenset([span])
        mappings = deepcopy(self.id_to_mappings.get(region_id, set()))
        return Entity(
            match=self.text[span.start : span.end],
            entity_class=label,
            spans=single_span,
            namespace="gold",
            mappings=mappings,
        )

    @staticmethod
    def _get_first_part_of_doc_id(task: dict[str, Any]) -> str:
        id_: str = task["data"]["id"]
        return id_.split("_")[0]

    @classmethod
    def convert_tasks_to_docs(cls, tasks: list[dict]) -> list[Document]:
        # group by first part of doc ID to get sections of multi part docs
        by_doc_id = sort_then_group(tasks, key_func=cls._get_first_part_of_doc_id)
        docs = []
        for doc_id, tasks_iter in by_doc_id:
            doc = Document(idx=doc_id)
            by_section = sort_then_group(
                tasks_iter, key_func=lambda x: tuple(x["data"]["id"].split("_")[1:])
            )
            for _, section_tasks_iter in by_section:
                for section_task in section_tasks_iter:
                    converter = LSToKazuConversion(section_task)
                    section = converter.create_section()
                    doc.sections.append(section)
            docs.append(doc)

        return docs


class LabelStudioAnnotationView:
    def __init__(self, ner_labels: dict[str, str]):
        """

        :param ner_labels: a mapping of ner label (i.e. :attr:`.Entity.entity_class`) to a valid colour
        """

        self.ner_labels = ner_labels

    @staticmethod
    def getDOM() -> XMLDocument:
        """
        ..
           (comment about why we've written this docstring like this for Sphinx)
           Explicit link with rtype because it's tricky to get the type hint to automatically:
           1. link to the python library docs
           2. have the link title make it obvious where the class is

        :rtype: :external+python:ref:`xml.dom.minidom.Document <dom-document-objects>`
        """
        impl = getDOMImplementation()
        if not isinstance(impl, DOMImplementation):
            raise RuntimeError("failed to get DOMImplementation")
        doc = impl.createDocument(
            None,
            "View",
            None,
        )
        if not isinstance(doc, XMLDocument):
            raise RuntimeError("failed to create document")
        return doc

    def build_labels(self, dom: XMLDocument, element: Element) -> None:
        """.. (sphinx comment) as above for why we have explicit type links.

        :type dom: :external+python:ref:`xml.dom.minidom.Document <dom-document-objects>`
        :type element: :external+python:ref:`xml.dom.minidom.Element <dom-element-objects>`
        """
        labels = dom.createElement("Labels")
        labels.setAttribute("name", "ner")
        labels.setAttribute("toName", "text")
        labels.setAttribute("choice", "multiple")
        for k, v in self.ner_labels.items():
            label = dom.createElement("Label")
            label.setAttribute("value", k)
            label.setAttribute("background", v)
            labels.appendChild(label)
        element.appendChild(labels)

    @staticmethod
    def build_taxonomy(dom: XMLDocument, element: Element, tasks: list[dict], name: str) -> None:
        """.. (sphinx comment) as above for why we have explicit type links.

        :type dom: :external+python:ref:`xml.dom.minidom.Document <dom-document-objects>`
        :type element: :external+python:ref:`xml.dom.minidom.Element <dom-element-objects>`
        """
        taxonomy = dom.createElement("Taxonomy")
        element.appendChild(taxonomy)
        taxonomy.setAttribute("visibleWhen", "region-selected")
        taxonomy.setAttribute("name", name)
        taxonomy.setAttribute("toName", "text")
        taxonomy.setAttribute("perRegion", "true")
        source_to_choice_element: dict[str, Element] = {}
        added: defaultdict[str, set[tuple[str, str]]] = defaultdict(set)
        for task in tasks:
            for annotation in task["annotations"]:
                for result in annotation["result"]:
                    tax_data = result.get("value", {}).get("taxonomy")
                    if tax_data is not None:
                        for tax in tax_data:
                            if len(tax) != 2:
                                logger.warning(
                                    f"taxonomy data malformed: {task['data']['id']}, {tax}"
                                )
                            else:
                                source, label_and_idx = tax[0], tax[1]
                                if label_and_idx in added[source]:
                                    continue
                                source_choice = source_to_choice_element.get(source)
                                if source_choice is None:
                                    source_choice = dom.createElement("Choice")
                                    source_choice.setAttribute("value", source)
                                    taxonomy.appendChild(source_choice)
                                    source_to_choice_element[source] = source_choice

                                choice = dom.createElement("Choice")
                                choice.setAttribute("value", label_and_idx)
                                source_choice.appendChild(choice)
                                added[source].add(label_and_idx)

    def create_main_view(self, tasks: list[dict]) -> str:
        dom = self.getDOM()
        root = dom.documentElement
        root.setAttribute("style", "display: flex;")
        parent_view = dom.createElement("View")
        root.appendChild(parent_view)
        relations_view = dom.createElement("View")
        parent_view.appendChild(relations_view)
        relations = dom.createElement("Relations")
        relations_view.appendChild(relations)
        relation = dom.createElement("Relation")
        relation.setAttribute("value", "non-contig")
        relations.appendChild(relation)
        self.build_labels(dom, relations_view)
        text_view = dom.createElement("View")
        parent_view.appendChild(text_view)
        text = dom.createElement("Text")
        text.setAttribute("name", "text")
        text.setAttribute("value", "$text")
        text_view.appendChild(text)
        taxonomy_view = dom.createElement("View")
        parent_view.appendChild(taxonomy_view)
        self.build_taxonomy(dom, taxonomy_view, tasks, _TAX_NAME)
        header_view = dom.createElement("View")
        parent_view.appendChild(header_view)
        header_view.setAttribute("style", "width: 100%; display: block")
        header = dom.createElement("Header")
        header.setAttribute("value", "Select span after creation to go next")
        header_view.appendChild(header)
        # this is a string when encoding is None.
        # see:
        # https://docs.python.org/3.11/library/xml.dom.minidom.html?highlight=minidom#xml.dom.minidom.Node.toxml
        dom_xml: str = dom.toxml()
        return dom_xml

    @staticmethod
    def with_default_colours() -> "LabelStudioAnnotationView":
        return LabelStudioAnnotationView(
            ner_labels={
                "cell_line": "red",
                "cell_type": "darkblue",
                "disease": "orange",
                "drug": "yellow",
                "gene": "green",
                "species": "purple",
                "anatomy": "pink",
                "go_mf": "grey",
                "go_cc": "blue",
                "go_bp": "brown",
            }
        )


class LabelStudioManager:
    def __init__(
        self,
        project_name: str,
        # headers could actually be Mapping[Union[str, bytes], Union[str, bytes]]
        # but typing-requests doesn't allow this.
        headers: Optional[CollectionsMapping[str, Union[str, bytes]]],
        url: str = "http://localhost:8080",
    ):
        self.project_name = project_name
        self.headers = headers
        self.url = url

    @cached_property
    def project_id(self) -> int:
        projects_url = f"{self.url}/api/projects"
        projects_resp = requests.get(projects_url, headers=self.headers)

        if not projects_resp.ok:
            raise HTTPError(
                f"failed to get projects from {projects_url}: status:{projects_resp.status_code}\n"
                f"{projects_resp.text}",
                response=projects_resp,
            )
        else:

            project_ids = [
                result
                for result in projects_resp.json()["results"]
                if result["title"] == self.project_name
            ]
            if len(project_ids) == 0:
                raise ValueError(f"no project with name: {self.project_name} found in Label Studio")
            elif len(project_ids) == 1:
                id_: int = project_ids[0]["id"]
                return id_
            else:
                raise ValueError(
                    f"more than one project with name: {self.project_name} found in Label Studio"
                )

    def delete_project_if_exists(self):
        try:
            resp = requests.delete(
                f"{self.url}/api/projects/{self.project_id}", headers=self.headers
            )
            if resp.status_code != 204:
                resp.raise_for_status()
        except (requests.exceptions.HTTPError, ValueError) as e:
            logger.warning(f"failed to delete project {self.project_name}. Maybe it doesn't exist?")
            logger.exception(e)
        if "project_id" in self.__dict__:
            del self.project_id

    def create_linking_project(self) -> None:
        payload = {"title": self.project_name}

        try:
            resp = requests.post(f"{self.url}/api/projects", json=payload, headers=self.headers)
            if resp.status_code != 201:
                resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logger.error(f"failed to create project {self.project_name}")
            raise e

    def update_view(self, view: LabelStudioAnnotationView, docs: list[Document]) -> None:
        tasks = KazuToLabelStudioConverter.convert_docs_to_tasks(docs)
        payload = {"label_config": view.create_main_view(tasks)}

        try:
            resp = requests.patch(
                f"{self.url}/api/projects/{self.project_id}", json=payload, headers=self.headers
            )
            if resp.status_code not in {200, 201}:
                resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logger.error(f"failed to update view for project {self.project_name}")
            raise e

    def update_tasks(self, docs: list[Document]) -> None:
        tasks = KazuToLabelStudioConverter.convert_docs_to_tasks(docs)
        try:
            resp = requests.post(
                f"{self.url}/api/projects/{self.project_id}/import",
                json=tasks,
                headers=self.headers,
            )
            if resp.status_code not in {200, 201}:
                resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logger.error(f"failed to update tasks for project {self.project_name}")
            raise e

    def get_all_tasks(self) -> list[dict[str, Any]]:
        tasks = requests.get(
            f"{self.url}/api/tasks?page_size=1000000&project={self.project_id}",
            headers=self.headers,
        ).json()
        ids = [task["id"] for task in tasks["tasks"]]
        return self.get_tasks(ids)

    def get_tasks(self, ids: list[int]) -> list[dict[str, Any]]:
        task_data = []
        for idx in ids:
            task_data.append(
                requests.get(
                    f"{self.url}/api/tasks/{idx}?project={self.project_id}",
                    headers=self.headers,
                ).json()
            )
        return task_data

    def export_from_ls(self) -> list[Document]:
        tasks = self.get_all_tasks()
        return LSToKazuConversion.convert_tasks_to_docs(tasks)



"""This module consists entirely of implementations of
:class:`~.OntologyParser`.

Some of these are aimed specifically at a custom format for individual
ontologies, like :class:`~.ChemblOntologyParser` or
:class:`~.MeddraOntologyParser`.

Others are aimed to provide flexibly for a user across a format, such as
:class:`~.RDFGraphParser`, :class:`~.TabularOntologyParser` and
:class:`~.JsonLinesOntologyParser`.

If you do not find a parser that meets your needs, please see
:ref:`writing-a-custom-parser`.
"""

import json
import logging
import os
import re
import sqlite3
from functools import cache
from pathlib import Path
from typing import cast, Any, Optional, Union, overload
from collections.abc import Iterable
from urllib import parse

import pandas as pd
import rdflib
from kazu.database.in_memory_db import MetadataDatabase

from kazu.data.data import (
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    CuratedTerm,
    AssociatedIdSets,
    GlobalParserActions,
)
from kazu.language.string_similarity_scorers import StringSimilarityScorer
from kazu.ontology_preprocessing.base import (
    OntologyParser,
    SYN,
    MAPPING_TYPE,
    DEFAULT_LABEL,
    IDX,
    IdsAndSource,
)
from kazu.ontology_preprocessing.synonym_generation import CombinatorialSynonymGenerator
from kazu.utils.utils import PathLike
from kazu.utils.grouping import sort_then_group

logger = logging.getLogger(__name__)


class JsonLinesOntologyParser(OntologyParser):
    """A parser for a jsonlines dataset.

    Assumes one kb entry per line (i.e. json object).

    This should be subclassed and subclasses must implement
    :meth:`~.json_dict_to_parser_records`.
    """

    def read(self, path: str) -> Iterable[dict[str, Any]]:
        for json_path in Path(path).glob("*.json"):
            with json_path.open(mode="r") as f:
                for line in f:
                    yield json.loads(line)

    def parse_to_dataframe(self):
        return pd.DataFrame.from_records(self.json_dict_to_parser_records(self.read(self.in_path)))

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[dict[str, Any]]
    ) -> Iterable[dict[str, Any]]:
        """For a given input json (represented as a python dict), yield
        dictionary record(s) compatible with the expected structure of the
        Ontology Parser superclass.

        This means dictionaries should have keys for :data:`~.SYN`,
        :data:`~.MAPPING_TYPE`, :data:`~.DEFAULT_LABEL` and
        :data:`~.IDX`. All other keys are used as mapping metadata.

        :param jsons_gen: iterable of python dict representing json objects
        :return:
        """
        raise NotImplementedError


class OpenTargetsDiseaseOntologyParser(JsonLinesOntologyParser):
    # Just use IDs that are in MONDO, since that's all people in general care about.
    # if we want to expand this out, other sources are:
    # "OGMS", "FBbt", "Orphanet", "EFO", "OTAR"
    # but we did have these in previously, and EFO introduced a lot of noise as it
    # has non-disease terms like 'dose' that occur frequently.
    # we could make the allowed sources a config option but we don't need to configure
    # currently, and easy to change later (and provide the current value as a default if
    # not present in config)
    allowed_sources = {"MONDO", "HP"}

    def find_kb(self, string: str) -> str:
        return string.split("_")[0]

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[dict[str, Any]]
    ) -> Iterable[dict[str, Any]]:
        # we ignore related syns for now until we decide how the system should handle them
        for json_dict in jsons_gen:
            idx = self.look_for_mondo(json_dict["id"], json_dict.get("dbXRefs", []))
            if any(allowed_source in idx for allowed_source in self.allowed_sources):
                synonyms = json_dict.get("synonyms", {})
                exact_syns = synonyms.get("hasExactSynonym", [])
                exact_syns.append(json_dict["name"])
                def_label = json_dict["name"]
                dbXRefs = json_dict.get("dbXRefs", [])
                for syn in exact_syns:
                    yield {
                        SYN: syn,
                        MAPPING_TYPE: "hasExactSynonym",
                        DEFAULT_LABEL: def_label,
                        IDX: idx,
                        "dbXRefs": dbXRefs,
                    }

    def look_for_mondo(self, ot_id: str, db_xrefs: list[str]) -> str:
        if "MONDO" in ot_id:
            return ot_id
        for x in db_xrefs:
            if "MONDO" in x:
                return x.replace(":", "_")
        return ot_id


class OpenTargetsTargetOntologyParser(JsonLinesOntologyParser):

    annotation_fields = {
        "subcellularLocations",
        "tractability",
        "constraint",
        "functionDescriptions",
        "go",
        "hallmarks",
        "chemicalProbes",
        "safetyLiabilities",
        "pathways",
        "targetClass",
    }

    def score_and_group_ids(
        self,
        ids_and_source: IdsAndSource,
        is_symbolic: bool,
    ) -> tuple[AssociatedIdSets, EquivalentIdAggregationStrategy]:
        """Group Ensembl gene IDs belonging to the same gene.

        .. admonition:: Note for non-biologists about genes
           :class: note

           The concept of a 'gene' is complex, and Ensembl gene IDs actually refer
           to locations on the genome, rather than individual genes. In fact, one
           'gene' can be made up of multiple Ensembl gene IDs, generally speaking
           these are exons that produce different isoforms of a given protein.

        :param ids_and_source:
        :param is_symbolic:
        :return:
        """
        meta_db = MetadataDatabase()
        assoc_id_sets: set[EquivalentIdSet] = set()

        for _, grouped_ids_and_source in sort_then_group(
            ids_and_source, lambda x: meta_db.get_by_idx(name=self.name, idx=x[0])[DEFAULT_LABEL]
        ):
            assoc_id_sets.add(
                EquivalentIdSet(
                    ids_and_source=frozenset(
                        (
                            grouped_idx,
                            source,
                        )
                        for grouped_idx, source in grouped_ids_and_source
                    )
                )
            )
        return frozenset(assoc_id_sets), EquivalentIdAggregationStrategy.CUSTOM

    def find_kb(self, string: str) -> str:
        return "ENSEMBL"

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[dict[str, Any]]
    ) -> Iterable[dict[str, Any]]:
        for json_dict in jsons_gen:
            # due to a bug in OT data, TEC genes have "gene" as a synonym. Since they're uninteresting, we just filter
            # them
            biotype = json_dict.get("biotype")
            if biotype == "" or biotype == "tec" or json_dict["id"] == json_dict["approvedSymbol"]:
                continue

            annotation_score = sum(
                1
                for annotation_field in self.annotation_fields
                if len(json_dict.get(annotation_field, [])) > 0
            )

            idx = json_dict["id"]
            default_label = json_dict["approvedSymbol"]
            shared_values = {
                IDX: idx,
                DEFAULT_LABEL: default_label,
                "dbXRefs": json_dict.get("dbXRefs", []),
                "approvedName": json_dict["approvedName"],
                "annotation_score": annotation_score,
            }

            for key in ["synonyms", "obsoleteSymbols", "obsoleteNames", "proteinIds"]:
                synonyms_and_sources_lst = json_dict.get(key, [])
                for record in synonyms_and_sources_lst:
                    if "label" in record and "id" in record:
                        raise RuntimeError(f"record: {record} has both id and label specified")
                    elif "label" in record:
                        record[SYN] = record.pop("label")
                    elif "id" in record:
                        record[SYN] = record.pop("id")
                    record[MAPPING_TYPE] = record.pop("source")
                    record.update(shared_values)
                    yield record

            for key in ("approvedSymbol", "approvedName", "id"):
                if key == "id":
                    mapping_type = "opentargets_id"
                else:
                    mapping_type = key

                res = {SYN: json_dict[key], MAPPING_TYPE: mapping_type}
                res.update(shared_values)
                yield res


class OpenTargetsMoleculeOntologyParser(JsonLinesOntologyParser):
    def find_kb(self, string: str) -> str:
        return "CHEMBL"

    def json_dict_to_parser_records(
        self, jsons_gen: Iterable[dict[str, Any]]
    ) -> Iterable[dict[str, Any]]:
        for json_dict in jsons_gen:
            cross_references = json_dict.get("crossReferences", {})
            default_label = json_dict["name"]
            idx = json_dict["id"]

            synonyms = json_dict.get("synonyms", [])
            main_name = json_dict["name"]
            synonyms.append(main_name)

            for syn in synonyms:
                yield {
                    SYN: syn,
                    MAPPING_TYPE: "synonyms",
                    "crossReferences": cross_references,
                    DEFAULT_LABEL: default_label,
                    IDX: idx,
                }

            for trade_name in json_dict.get("tradeNames", []):
                yield {
                    SYN: trade_name,
                    MAPPING_TYPE: "tradeNames",
                    "crossReferences": cross_references,
                    DEFAULT_LABEL: default_label,
                    IDX: idx,
                }


RdfRef = Union[rdflib.paths.Path, rdflib.term.Node, str]
# Note - lists are actually normally provided here through hydra config
# but there's apparently no way of type hinting
# 'any iterable of length two where the items have these types'
PredicateAndValue = tuple[RdfRef, rdflib.term.Node]


class RDFGraphParser(OntologyParser):
    """Parser for rdf files.

    Supports any format of rdf that :meth:`rdflib.Graph.parse` can infer
    the format of from the file extension, e.g. ``.xml`` , ``.ttl`` ,
    ``.owl`` , ``.json``. Case of the extension does not matter. This
    functionality is handled by :func:`rdflib.util.guess_format`, but
    will fall back to attempting to parse as turtle/ttl format in the
    case of an unknown file extension.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        uri_regex: Union[str, re.Pattern],
        synonym_predicates: Iterable[RdfRef],
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        label_predicate: RdfRef = rdflib.RDFS.label,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )

        if isinstance(uri_regex, re.Pattern):
            self._uri_regex = uri_regex
        else:
            self._uri_regex = re.compile(uri_regex)

        self.synonym_predicates = tuple(
            self.convert_to_rdflib_ref(pred) for pred in synonym_predicates
        )
        self.label_predicate = self.convert_to_rdflib_ref(label_predicate)

        if include_entity_patterns is not None:
            self.include_entity_patterns = tuple(
                (self.convert_to_rdflib_ref(pred), self.convert_to_rdflib_ref(val))
                for pred, val in include_entity_patterns
            )
        else:
            self.include_entity_patterns = tuple()

        if exclude_entity_patterns is not None:
            self.exclude_entity_patterns = tuple(
                (self.convert_to_rdflib_ref(pred), self.convert_to_rdflib_ref(val))
                for pred, val in exclude_entity_patterns
            )
        else:
            self.exclude_entity_patterns = tuple()

    def find_kb(self, string: str) -> str:
        """By default, just return the name of the parser.

        If more complex behaviour is necessary, write a custom subclass
        and override this method.
        """
        return self.name

    @overload
    @staticmethod
    def convert_to_rdflib_ref(pred: rdflib.paths.Path) -> rdflib.paths.Path:
        pass

    @overload
    @staticmethod
    def convert_to_rdflib_ref(pred: rdflib.term.Node) -> rdflib.term.Node:
        pass

    @overload
    @staticmethod
    def convert_to_rdflib_ref(pred: str) -> rdflib.URIRef:
        pass

    @staticmethod
    def convert_to_rdflib_ref(pred):
        if isinstance(pred, (rdflib.term.Node, rdflib.paths.Path)):
            return pred
        else:
            return rdflib.URIRef(pred)

    def parse_to_dataframe(self) -> pd.DataFrame:
        g = rdflib.Graph()
        g.parse(self.in_path)
        default_labels = []
        iris = []
        syns = []
        mapping_type = []

        label_pred_str = str(self.label_predicate)

        for sub, obj in g.subject_objects(self.label_predicate):
            if not self.is_valid_iri(str(sub)):
                continue

            if any((sub, pred, value) not in g for pred, value in self.include_entity_patterns):
                continue

            if any((sub, pred, value) in g for pred, value in self.exclude_entity_patterns):
                continue

            default_labels.append(str(obj))
            iris.append(str(sub))
            syns.append(str(obj))
            mapping_type.append(label_pred_str)
            for syn_predicate in self.synonym_predicates:
                for other_syn_obj in g.objects(subject=sub, predicate=syn_predicate):
                    default_labels.append(str(obj))
                    iris.append(str(sub))
                    syns.append(str(other_syn_obj))
                    mapping_type.append(str(syn_predicate))

        df = pd.DataFrame.from_dict(
            {DEFAULT_LABEL: default_labels, IDX: iris, SYN: syns, MAPPING_TYPE: mapping_type}
        )
        return df

    def is_valid_iri(self, text: str) -> bool:
        """Check if input string is a valid IRI for the ontology being parsed.

        Uses ``self._uri_regex`` to define valid IRIs.
        """
        match = self._uri_regex.match(text)
        return bool(match)


SKOS_XL_PREF_LABEL_PATH: rdflib.paths.Path = rdflib.URIRef(
    "http://www.w3.org/2008/05/skos-xl#prefLabel"
) / rdflib.URIRef("http://www.w3.org/2008/05/skos-xl#literalForm")
SKOS_XL_ALT_LABEL_PATH: rdflib.paths.Path = rdflib.URIRef(
    "http://www.w3.org/2008/05/skos-xl#altLabel"
) / rdflib.URIRef("http://www.w3.org/2008/05/skos-xl#literalForm")


class SKOSXLGraphParser(RDFGraphParser):
    """Parse SKOS-XL RDF Files.

    Note that this just sets a default label predicate and synonym
    predicate to SKOS-XL appropriate paths, and then passes to the
    parent RDFGraphParser class. This class is just a convenience to
    make specifying a SKOS-XL parser easier, this functionality is still
    available via RDFGraphParser directly.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        uri_regex: Union[str, re.Pattern],
        synonym_predicates: Iterable[RdfRef] = (SKOS_XL_ALT_LABEL_PATH,),
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        label_predicate: RdfRef = SKOS_XL_PREF_LABEL_PATH,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=uri_regex,
            synonym_predicates=synonym_predicates,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            include_entity_patterns=include_entity_patterns,
            exclude_entity_patterns=exclude_entity_patterns,
            curations_path=curations_path,
            global_actions=global_actions,
            label_predicate=label_predicate,
        )


class GeneOntologyParser(OntologyParser):
    _uri_regex = re.compile("^http://purl.obolibrary.org/obo/GO_[0-9]+$")

    instances: set[str] = set()
    instances_in_dbs: set[str] = set()

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        query: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )
        self.instances.add(name)
        self.query = query

    def populate_databases(
        self, force: bool = False, return_curations: bool = False
    ) -> Optional[list[CuratedTerm]]:
        curations = super().populate_databases(force=force, return_curations=return_curations)
        self.instances_in_dbs.add(self.name)

        if self.instances_in_dbs >= self.instances:
            # all existing instances are in the database, so we can free up
            # the memory used by the cached parsed gene ontology, which is significant.
            self.load_go.cache_clear()
        return curations

    def __del__(self):
        GeneOntologyParser.instances.discard(self.name)

    @staticmethod
    @cache
    def load_go(in_path: PathLike) -> rdflib.Graph:
        g = rdflib.Graph()
        g.parse(in_path)
        return g

    def find_kb(self, string: str) -> str:
        return self.name

    def parse_to_dataframe(self) -> pd.DataFrame:
        g = self.load_go(self.in_path)
        result = g.query(self.query)
        default_labels = []
        iris = []
        syns = []
        mapping_type = []

        # there seems to be a bug in rdflib that means the iterator sometimes exits early unless we convert to list first
        # type cast is necessary because iterating over an rdflib query result gives different types depending on the kind
        # of query, so rdflib gives a Union here, but we know it should be a ResultRow because we know we should have a
        # select query
        list_res = cast(list[rdflib.query.ResultRow], list(result))
        for row in list_res:
            idx = str(row.goid)
            label = str(row.label)
            if "obsolete" in label:
                logger.debug("skipping obsolete id: %s, %s", idx, label)
                continue
            if self._uri_regex.match(idx):
                default_labels.append(label)
                iris.append(idx)
                syns.append(str(row.synonym))
                mapping_type.append("hasExactSynonym")
        df = pd.DataFrame.from_dict(
            {DEFAULT_LABEL: default_labels, IDX: iris, SYN: syns, MAPPING_TYPE: mapping_type}
        )
        default_labels_df = df[[IDX, DEFAULT_LABEL]].drop_duplicates().copy()
        default_labels_df[SYN] = default_labels_df[DEFAULT_LABEL]
        default_labels_df[MAPPING_TYPE] = "label"

        return pd.concat([df, default_labels_df])


class BiologicalProcessGeneOntologyParser(GeneOntologyParser):
    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
            query="""
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                    PREFIX oboinowl: <http://www.geneontology.org/formats/oboInOwl#>

                    SELECT DISTINCT ?goid ?label ?synonym
                            WHERE {

                                ?goid oboinowl:hasExactSynonym ?synonym .
                                ?goid rdfs:label ?label .
                                ?goid oboinowl:hasOBONamespace "biological_process" .

                  }
            """,
        )


class MolecularFunctionGeneOntologyParser(GeneOntologyParser):
    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
            query="""
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                    PREFIX oboinowl: <http://www.geneontology.org/formats/oboInOwl#>

                    SELECT DISTINCT ?goid ?label ?synonym
                            WHERE {

                                ?goid oboinowl:hasExactSynonym ?synonym .
                                ?goid rdfs:label ?label .
                                ?goid oboinowl:hasOBONamespace "molecular_function".

                    }
            """,
        )


class CellularComponentGeneOntologyParser(GeneOntologyParser):
    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
            query="""
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
                    PREFIX oboinowl: <http://www.geneontology.org/formats/oboInOwl#>

                    SELECT DISTINCT ?goid ?label ?synonym
                            WHERE {

                                ?goid oboinowl:hasExactSynonym ?synonym .
                                ?goid rdfs:label ?label .
                                ?goid oboinowl:hasOBONamespace "cellular_component" .

                    }
            """,
        )


class UberonOntologyParser(RDFGraphParser):
    """Input should be an UBERON owl file e.g.
    https://www.ebi.ac.uk/ols/ontologies/uberon."""

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):

        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/UBERON_[0-9]+$"),
            synonym_predicates=(
                rdflib.URIRef("http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"),
            ),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "UBERON"


class MondoOntologyParser(OntologyParser):
    _uri_regex = re.compile("^http://purl.obolibrary.org/obo/(MONDO|HP)_[0-9]+$")
    """Input should be a MONDO json file e.g.
    https://www.ebi.ac.uk/ols/ontologies/mondo."""

    def find_kb(self, string: str) -> str:
        path = parse.urlparse(string).path
        # just the final bit, e.g. MONDO_0000123
        path_end = path.split("/")[-1]
        # we don't want the underscore or digits for the unique ID, just the ontology bit
        return path_end.split("_")[0]

    def parse_to_dataframe(self) -> pd.DataFrame:
        x = json.load(open(self.in_path, "r"))
        graph = x["graphs"][0]
        nodes = graph["nodes"]
        ids = []
        default_label_list = []
        all_syns = []
        mapping_type = []
        for i, node in enumerate(nodes):
            if not self.is_valid_iri(node["id"]):
                continue

            idx = node["id"]
            default_label = node.get("lbl")
            if default_label is None:
                # skip if no default label is available
                continue
            # add default_label to syn type
            all_syns.append(default_label)
            default_label_list.append(default_label)
            mapping_type.append("lbl")
            ids.append(idx)

            syns = node.get("meta", {}).get("synonyms", [])
            for syn_dict in syns:

                pred = syn_dict["pred"]
                if pred in {"hasExactSynonym"}:
                    mapping_type.append(pred)
                    syn = syn_dict["val"]
                    ids.append(idx)
                    default_label_list.append(default_label)
                    all_syns.append(syn)

        df = pd.DataFrame.from_dict(
            {IDX: ids, DEFAULT_LABEL: default_label_list, SYN: all_syns, MAPPING_TYPE: mapping_type}
        )
        return df

    def is_valid_iri(self, text: str) -> bool:
        match = self._uri_regex.match(text)
        return bool(match)


class HGNCGeneOntologyParser(OntologyParser):
    """Parse HGNC data and extract individual genes as entities.

    Input is a json from HGNC. For example,
    http://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/json/hgnc_complete_set.json.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            name=name,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "ENSEMBL"

    def parse_to_dataframe(self) -> pd.DataFrame:

        keys_to_check = [
            "name",
            "symbol",
            "uniprot_ids",
            "alias_name",
            "alias_symbol",
            "prev_name",
            "lncipedia",
            "prev_symbol",
            "vega_id",
            "refseq_accession",
            "hgnc_id",
            "mgd_id",
            "rgd_id",
            "ccds_id",
            "pseudogene.org",
        ]

        with open(self.in_path, "r") as f:
            data = json.load(f)
        ids = []
        default_label = []
        all_syns = []
        all_mapping_type: list[str] = []
        docs = data["response"]["docs"]
        for doc in docs:

            def get_with_default_list(key: str) -> list[str]:
                found = doc.get(key, [])
                if not isinstance(found, list):
                    found = [found]
                return cast(list[str], found)

            ensembl_gene_id = doc.get("ensembl_gene_id", None)
            name = doc.get("name", None)
            if ensembl_gene_id is None or name is None:
                continue
            else:
                # find synonyms
                synonyms: list[tuple[str, str]] = []
                for hgnc_key in keys_to_check:
                    synonyms_this_entity = get_with_default_list(hgnc_key)
                    for potential_synonym in synonyms_this_entity:
                        synonyms.append((potential_synonym, hgnc_key))

                synonyms = list(set(synonyms))
                synonyms_strings = []
                for synonym_str, mapping_t in synonyms:
                    all_mapping_type.append(mapping_t)
                    synonyms_strings.append(synonym_str)

                num_syns = len(synonyms_strings)
                ids.extend([ensembl_gene_id] * num_syns)
                default_label.extend([name] * num_syns)
                all_syns.extend(synonyms_strings)

        df = pd.DataFrame.from_dict(
            {IDX: ids, DEFAULT_LABEL: default_label, SYN: all_syns, MAPPING_TYPE: all_mapping_type}
        )
        return df


class ChemblOntologyParser(OntologyParser):
    """Input is a directory containing an extracted sqllite dump from Chembl.

    For example, this can be sourced from:
    https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_33/chembl_33_sqlite.tar.gz.
    """

    def find_kb(self, string: str) -> str:
        return "CHEMBL"

    def parse_to_dataframe(self) -> pd.DataFrame:
        conn = sqlite3.connect(self.in_path)
        query = f"""\
            SELECT chembl_id AS {IDX}, pref_name AS {DEFAULT_LABEL}, synonyms AS {SYN}, syn_type AS {MAPPING_TYPE}
            FROM molecule_dictionary AS md
                     JOIN molecule_synonyms ms ON md.molregno = ms.molregno
            UNION ALL
            SELECT chembl_id AS {IDX}, pref_name AS {DEFAULT_LABEL}, pref_name AS {SYN}, 'pref_name' AS {MAPPING_TYPE}
            FROM molecule_dictionary
        """
        df = pd.read_sql(query, conn)
        # eliminate anything without a pref_name, as will be too big otherwise
        df = df.dropna(subset=[DEFAULT_LABEL])

        df.drop_duplicates(inplace=True)

        return df


class CLOOntologyParser(RDFGraphParser):
    """Input is a CLO Owl file https://www.ebi.ac.uk/ols/ontologies/clo."""

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/CLO_[0-9]+$"),
            synonym_predicates=(
                rdflib.URIRef("http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"),
            ),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "CLO"


class CellosaurusOntologyParser(OntologyParser):
    """Input is an obo file from cellosaurus, e.g.
    https://ftp.expasy.org/databases/cellosaurus/cellosaurus.obo."""

    cell_line_re = re.compile("cell line", re.IGNORECASE)

    def find_kb(self, string: str) -> str:
        return "CELLOSAURUS"

    def score_and_group_ids(
        self,
        ids_and_source: IdsAndSource,
        is_symbolic: bool,
    ) -> tuple[AssociatedIdSets, EquivalentIdAggregationStrategy]:
        """Treat all synonyms as seperate cell lines.

        :param ids_and_source:
        :param is_symbolic:
        :return:
        """

        return (
            frozenset(
                EquivalentIdSet(
                    ids_and_source=frozenset((single_id_and_source,)),
                )
                for single_id_and_source in ids_and_source
            ),
            EquivalentIdAggregationStrategy.CUSTOM,
        )

    def _remove_cell_line_text(self, text: str) -> str:
        return self.cell_line_re.sub("", text).strip()

    def parse_to_dataframe(self) -> pd.DataFrame:

        ids = []
        default_labels = []
        all_syns = []
        mapping_type = []
        with open(self.in_path, "r") as f:
            id = ""
            for line in f:
                text = line.rstrip()
                if text.startswith("id:"):
                    id = text.split(" ")[1]
                elif text.startswith("name:"):
                    default_label = text[5:].strip()
                    ids.append(id)
                    # we remove "cell line" because they're all cell lines and it confuses mapping
                    default_label_no_cell_line = self._remove_cell_line_text(default_label)
                    default_labels.append(default_label_no_cell_line)
                    all_syns.append(default_label_no_cell_line)
                    mapping_type.append("name")
                # synonyms in cellosaurus are a bit of a mess, so we don't use this field for now. Leaving this here
                # in case they improve at some point
                # elif text.startswith("synonym:"):
                #     match = self._synonym_regex.match(text)
                #     if match is None:
                #         raise ValueError(
                #             """synonym line does not match our synonym regex.
                #             Either something is wrong with the file, or it has updated
                #             and our regex is not correct/general enough."""
                #         )
                #     ids.append(id)
                #     default_labels.append(default_label)
                #
                #     all_syns.append(self._remove_cell_line_text(match.group("syn")))
                #     mapping_type.append(match.group("mapping"))
                else:
                    pass
        df = pd.DataFrame.from_dict(
            {IDX: ids, DEFAULT_LABEL: default_labels, SYN: all_syns, MAPPING_TYPE: mapping_type}
        )
        return df

    _synonym_regex = re.compile(
        r"""^synonym:      # line that begins synonyms
        \s*                # any amount of whitespace (standardly a single space)
        "(?P<syn>[^"]*)"   # a quoted string - capture this as a named match group 'syn'
        \s*                # any amount of separating whitespace (standardly a single space)
        (?P<mapping>\w*)   # a sequence of word characters representing the mapping type
        \s*                # any amount of separating whitespace (standardly a single space)
        \[\]               # an open and close bracket at the end of the string
        $""",
        re.VERBOSE,
    )


class MeddraOntologyParser(OntologyParser):
    """input is an unzipped directory to a Meddra release (Note, requires
    licence).

    This should contain the files 'mdhier.asc' and 'llt.asc'.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        exclude_socs: Iterable[str] = (
            "Surgical and medical procedures",
            "Social circumstances",
            "Investigations",
        ),
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
            name=name,
        )

        self.exclude_socs = exclude_socs

    _mdhier_asc_col_names = (
        "pt_code",
        "hlt_code",
        "hlgt_code",
        "soc_code",
        "pt_name",
        "hlt_name",
        "hlgt_name",
        "soc_name",
        "soc_abbrev",
        "null_field",
        "pt_soc_code",
        "primary_soc_fg",
        "NULL",
    )

    _llt_asc_column_names = (
        "llt_code",
        "llt_name",
        "pt_code",
        "llt_whoart_code",
        "llt_harts_code",
        "llt_costart_sym",
        "llt_icd9_code",
        "llt_icd9cm_code",
        "llt_icd10_code",
        "llt_currency",
        "llt_jart_code",
        "NULL",
    )

    def find_kb(self, string: str) -> str:
        return "MEDDRA"

    def parse_to_dataframe(self) -> pd.DataFrame:
        # hierarchy path
        mdheir_path = os.path.join(self.in_path, "mdhier.asc")
        # low level term path
        llt_path = os.path.join(self.in_path, "llt.asc")
        hier_df = pd.read_csv(
            mdheir_path,
            sep="$",
            header=None,
            names=self._mdhier_asc_col_names,
            dtype="string",
        )
        hier_df = hier_df[~hier_df["soc_name"].isin(self.exclude_socs)]

        llt_df = pd.read_csv(
            llt_path,
            sep="$",
            header=None,
            names=self._llt_asc_column_names,
            usecols=("llt_name", "pt_code"),
            dtype="string",
        )
        llt_df = llt_df.dropna(axis=1)

        ids = []
        default_labels = []
        all_syns = []
        mapping_type = []
        soc_names = []
        soc_codes = []

        for i, row in hier_df.iterrows():
            idx = row["pt_code"]
            pt_name = row["pt_name"]
            soc_name = row["soc_name"]
            soc_code = row["soc_code"]
            llts = llt_df[llt_df["pt_code"] == idx]
            ids.append(idx)
            default_labels.append(pt_name)
            all_syns.append(pt_name)
            soc_names.append(soc_name)
            soc_codes.append(soc_code)
            mapping_type.append("meddra_link")
            for j, llt_row in llts.iterrows():
                ids.append(idx)
                default_labels.append(pt_name)
                soc_names.append(soc_name)
                soc_codes.append(soc_code)
                all_syns.append(llt_row["llt_name"])
                mapping_type.append("meddra_link")

        for i, row in (
            hier_df[["hlt_code", "hlt_name", "soc_name", "soc_code"]].drop_duplicates().iterrows()
        ):
            ids.append(row["hlt_code"])
            default_labels.append(row["hlt_name"])
            soc_names.append(row["soc_name"])
            soc_codes.append(row["soc_code"])
            all_syns.append(row["hlt_name"])
            mapping_type.append("meddra_link")
        for i, row in (
            hier_df[["hlgt_code", "hlgt_name", "soc_name", "soc_code"]].drop_duplicates().iterrows()
        ):
            ids.append(row["hlgt_code"])
            default_labels.append(row["hlgt_name"])
            soc_names.append(row["soc_name"])
            soc_codes.append(row["soc_code"])
            all_syns.append(row["hlgt_name"])
            mapping_type.append("meddra_link")
        df = pd.DataFrame.from_dict(
            {
                IDX: ids,
                DEFAULT_LABEL: default_labels,
                SYN: all_syns,
                MAPPING_TYPE: mapping_type,
                "soc_name": soc_names,
                "soc_code": soc_codes,
            }
        )
        return df


class CLOntologyParser(RDFGraphParser):
    """Input should be an CL owl file e.g.
    https://www.ebi.ac.uk/ols/ontologies/cl."""

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):

        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/CL_[0-9]+$"),
            synonym_predicates=(
                rdflib.URIRef("http://www.geneontology.org/formats/oboInOwl#hasExactSynonym"),
            ),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            include_entity_patterns=include_entity_patterns,
            exclude_entity_patterns=exclude_entity_patterns,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "CL"


class HGNCGeneFamilyParser(OntologyParser):
    """Parse HGNC data and extract only Gene Families as entities.

    Input is a json from HGNC. For example,
    http://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/json/hgnc_complete_set.json.
    """

    syn_column_keys = {"Family alias", "Common root gene symbol"}

    def find_kb(self, string: str) -> str:
        return "HGNC_GENE_FAMILY"

    def parse_to_dataframe(self) -> pd.DataFrame:
        df = pd.read_csv(self.in_path, sep="\t")
        data = []
        for family_id, row in (
            df.groupby(by="Family ID").agg(lambda col_series: set(col_series.dropna())).iterrows()
        ):
            # in theory, there should only be one family name per ID
            assert len(row["Family name"]) == 1
            default_label = next(iter(row["Family name"]))
            data.append(
                {
                    SYN: default_label,
                    MAPPING_TYPE: "Family name",
                    DEFAULT_LABEL: default_label,
                    IDX: family_id,
                }
            )
            data.extend(
                {
                    SYN: syn,
                    MAPPING_TYPE: key,
                    DEFAULT_LABEL: default_label,
                    IDX: family_id,
                }
                for key in self.syn_column_keys
                for syn in row[key]
            )
        return pd.DataFrame.from_records(data)


class TabularOntologyParser(OntologyParser):
    """For already tabulated data.

    This expects ``in_path`` to be the path to a file that can be loaded
    by :func:`pandas.read_csv` (e.g. a `.csv` or `.tsv` file),
    and the result be in the format that is produced by
    :meth:`~.OntologyParser.parse_to_dataframe` - see the docs of that
    method for more details on the format of this dataframe.

    Note that this class's ``__init__` method takes a ``kwargs`` param
    which is passed through to :func:`pandas.read_csv` , which gives you
    a notable degree of flexibility on how exactly the input file is
    converted into this dataframe. Although if this becomes complex to
    pass through in the ``kwargs``, it may be worth considering
    :ref:`writing-a-custom-parser`.
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
        **kwargs: Any,
    ):
        """

        :param in_path:
        :param entity_class:
        :param name:
        :param string_scorer:
        :param synonym_merge_threshold:
        :param data_origin:
        :param synonym_generator:
        :param curations_path:
        :param kwargs: passed to pandas.read_csv
        """
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
        )
        self._raw_dataframe: pd.DataFrame = pd.read_csv(self.in_path, **kwargs)

    def parse_to_dataframe(self) -> pd.DataFrame:
        """Assume input file is already in correct format.

        Inherit and override this method if different behaviour is required.

        :return:
        """
        return self._raw_dataframe

    def find_kb(self, string: str) -> str:
        return self.name


class ATCDrugClassificationParser(TabularOntologyParser):
    """Parser for the ATC Drug classification dataset.

    This requires a licence from WHO, available at
    https://www.who.int/tools/atc-ddd-toolkit/atc-classification
    .
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            curations_path=curations_path,
            global_actions=global_actions,
            sep="     ",
            header=None,
            names=["code", "level_and_description"],
            # Because the c engine can't handle multi-char sep
            # removing this results in the same behaviour, but
            # pandas logs a warning.
            engine="python",
        )

    levels_to_ignore = {"1", "2", "3"}

    def parse_to_dataframe(self) -> pd.DataFrame:
        # for some reason, the level and description codes are merged, so we need to fix this here
        df = self._raw_dataframe.applymap(str.strip)
        res_df = pd.DataFrame()
        res_df[[MAPPING_TYPE, DEFAULT_LABEL]] = df.apply(
            lambda row: [row["level_and_description"][0], row["level_and_description"][1:]],
            axis=1,
            result_type="expand",
        )
        res_df[IDX] = df["code"]
        res_df = res_df[~res_df[MAPPING_TYPE].isin(self.levels_to_ignore)]
        res_df[SYN] = res_df[DEFAULT_LABEL]
        return res_df


class StatoParser(RDFGraphParser):
    """Parse stato: input should be an owl file.

    Available at e.g.
    https://www.ebi.ac.uk/ols/ontologies/stato .
    """

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.7,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        include_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        exclude_entity_patterns: Optional[Iterable[PredicateAndValue]] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):

        super().__init__(
            in_path=in_path,
            entity_class=entity_class,
            name=name,
            uri_regex=re.compile("^http://purl.obolibrary.org/obo/(OBI|STATO)_[0-9]+$"),
            synonym_predicates=(rdflib.URIRef("http://purl.obolibrary.org/obo/IAO_0000111"),),
            string_scorer=string_scorer,
            synonym_merge_threshold=synonym_merge_threshold,
            data_origin=data_origin,
            synonym_generator=synonym_generator,
            include_entity_patterns=include_entity_patterns,
            exclude_entity_patterns=exclude_entity_patterns,
            curations_path=curations_path,
            global_actions=global_actions,
        )

    def find_kb(self, string: str) -> str:
        return "OBI" if "OBI" in string else "STATO"



"""Influenced by Huawei Noah's Ark Lab `TinyBERT <https://github.com/huawei-
noah/Pretrained-Language-Model/blob/master/TinyBERT>`_, but heavily modified
structurally to fit in our PyTorch Lightning training setup.

`This section of the TinyBERT code <https://github.com/huawei-noah/Pretrained-Language-Model/blob/master/TinyBERT/task_distill.py#L789-L1081>`_
in particular is relevant.

Licensed under Apache 2.0

| Copyright 2020 Huawei Technologies Co., Ltd.
| Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team.
| Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.

.. raw:: html

    <details>
    <summary>Full License Notice</summary>

| Copyright 2020 Huawei Technologies Co., Ltd.
| Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team.
| Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

.. raw:: html

    </details>
"""

import logging
from typing import Union, Optional, Any, cast
from collections.abc import Callable

import numpy as np
import pytorch_lightning as pl
import torch
from cachetools import LRUCache
from omegaconf import ListConfig, OmegaConf
from pytorch_lightning.utilities.types import TRAIN_DATALOADERS, EVAL_DATALOADERS
from torch.nn import CrossEntropyLoss, MSELoss
from torch.optim.lr_scheduler import LRScheduler
from torch.optim import AdamW
from torch.utils.data import Dataset, DataLoader
from transformers import (
    AutoTokenizer,
    InputExample,
    DataCollatorForTokenClassification,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
    get_constant_schedule,
    get_linear_schedule_with_warmup,
    get_cosine_schedule_with_warmup,
    get_constant_schedule_with_warmup,
)
from transformers.utils import check_min_version

from kazu.distillation.dataprocessor import NerProcessor
from kazu.distillation.metrics import numeric_label_f1_score, IGNORE_IDX
from kazu.distillation.tiny_transformers import TinyBertForSequenceTagging

check_min_version("4.0.0")  # at least 4.0.0... for optimerzers

logger = logging.getLogger(__name__)

SCHEDULES: dict[Optional[str], Callable] = {
    None: get_constant_schedule,
    "none": get_constant_schedule,
    "warmup_cosine": get_cosine_schedule_with_warmup,
    "warmup_constant": get_constant_schedule_with_warmup,
    "warmup_linear": get_linear_schedule_with_warmup,
    "torchStepLR": torch.optim.lr_scheduler.StepLR,
}


class NerDataset(Dataset):
    """A dataset used for Ner.

    designed for on the fly tokenisation to speed up multi processing.
    Uses caching to prevent repeated processing
    """

    def __init__(
        self,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        examples: list[InputExample],
        label_map: dict[str, int],
        max_length: int,
    ):
        """

        :param tokenizer: typically created from AutoTokenizer.from_pretrained
        :param examples: a list of InputExample, typically created from a
            :class:`kazu.distillation.dataprocessor.NerProcessor`
        :param label_map: str to int mapping of labels
        :param max_length: The maximum number of tokens per instance that the model can handle.
            Inputs longer than max_length value will be truncated.
        """
        self.label_map = label_map
        self.examples = examples
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.call_count = 0
        self.cache: LRUCache[int, dict[str, list]] = LRUCache(5000)

    def __getitem__(self, index: int) -> dict[str, list]:
        if index not in self.cache:
            self.cache[index] = self.convert_single_example(
                ex_index=index, example=self.examples[index]
            )
        self.call_count += 1
        return self.cache[index]

    def __len__(self):
        return len(self.examples)

    def convert_single_example(self, ex_index: int, example: InputExample) -> dict[str, list]:
        textlist = example.text_a.split()
        labellist = example.label.split()
        tokens: list[str] = []
        labels: list[str] = []
        for i, word in enumerate(textlist):
            tokenized = self.tokenizer.tokenize(word)
            tokens.extend(tokenized)
            label_1 = labellist[i]
            for m, tok in enumerate(tokenized):
                if m == 0:
                    labels.append(label_1)
                else:
                    labels.append("X")

        ntokens: list[str] = []
        segment_ids: list[int] = []
        label_id: list[int] = []
        ntokens.append("[CLS]")
        segment_ids.append(0)
        label_id.append(IGNORE_IDX)
        for i, token in enumerate(tokens):
            ntokens.append(token)
            segment_ids.append(0)
            if labels[i] == "X":
                label_id.append(IGNORE_IDX)
            else:
                label_id.append(self.label_map[labels[i]])

        # Truncation
        if len(ntokens) > self.max_length - 1:
            assert (len(ntokens) == len(segment_ids)) and (len(ntokens) == len(label_id))
            ntokens = ntokens[: self.max_length - 1]
            segment_ids = segment_ids[: self.max_length - 1]
            label_id = label_id[: self.max_length - 1]

        ntokens.append("[SEP]")
        segment_ids.append(0)
        label_id.append(IGNORE_IDX)

        input_ids = self.tokenizer.convert_tokens_to_ids(ntokens)
        # convert_tokens_to_ids can return a single int if a single token is passed,
        # but we know here we're dealing with lists.
        assert isinstance(input_ids, list)
        # The mask has 1 for real tokens and 0 for padding tokens.
        input_mask = [1] * len(input_ids)
        if self.call_count < 4 and ex_index < 4:  # Examples. Executed only once per model run
            logger.info("*** Example ***")
            logger.info("guid: %s" % (example.guid))
            logger.info("tokens: %s" % " ".join(tokens))
            logger.info("ntokens: %s" % " ".join(ntokens))
            logger.info("input_ids: %s" % " ".join(str(x) for x in input_ids))
            logger.info("input_mask: %s" % " ".join(str(x) for x in input_mask))
            logger.info("segment_ids: %s" % " ".join(str(x) for x in segment_ids))
            logger.info("label_id: %s" % " ".join(str(x) for x in label_id))

        result = {
            "labels": label_id,
            "input_ids": input_ids,
            "token_type_ids": segment_ids,
            "attention_mask": input_mask,
        }
        return result


class TaskSpecificDistillation(pl.LightningModule):
    def __init__(
        self,
        temperature: float,
        warmup_steps: int,
        learning_rate: float,
        weight_decay: float,
        batch_size: int,
        accumulate_grad_batches: int,
        max_epochs: int,
        schedule: Optional[str] = None,
    ):
        """Base class for distillation on PyTorch Lightning platform.

        :param temperature:
        :param warmup_steps:
        :param learning_rate:
        :param weight_decay:
        :param batch_size:
        :param accumulate_grad_batches:
        :param max_epochs:
        :param schedule:
        """

        super().__init__()
        self.accumulate_grad_batches = accumulate_grad_batches
        self.batch_size = batch_size
        self.weight_decay = weight_decay
        self.learning_rate = learning_rate

        self.temperature = temperature
        self.schedule = schedule
        self.training_examples = self.get_training_examples()

        self.num_train_optimization_steps = int(
            len(self.training_examples)
            / self.batch_size
            / self.accumulate_grad_batches
            * max_epochs
        )
        self.warmup_steps = warmup_steps
        logger.info(
            "num_train_optimization_steps: {}, args.warmup_steps: {}, args.gradient_accumulation_steps: {}".format(
                self.num_train_optimization_steps, self.warmup_steps, self.accumulate_grad_batches
            )
        )

    def get_training_examples(self) -> list[InputExample]:
        """Subclasses should implement this.

        :return:
        """
        raise NotImplementedError

    def get_optimizer_grouped_parameters(self, student_model):
        param_optimizer = list(student_model.named_parameters())
        size = 0
        logger.info("student_model.named_parameters :")
        for n, p in student_model.named_parameters():
            logger.info("n: {}".format(n))
            size += p.nelement()
        logger.info("Total parameters: {}".format(size))

        no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
                "weight_decay": self.weight_decay,
            },
            {
                "params": [p for n, p in param_optimizer if any(nd in n for nd in no_decay)],
                "weight_decay": 0.0,
            },
        ]
        return optimizer_grouped_parameters

    def configure_optimizers(
        self,
    ):
        """Configure optimizer and learning rate scheduler."""
        optimizer_grouped_parameters = self.get_optimizer_grouped_parameters(
            student_model=self.student_model
        )
        optimizer = AdamW(params=optimizer_grouped_parameters, lr=self.learning_rate)

        if self.schedule in ["torchStepLR"]:  # Multi-GPU : must use torch scheduler
            scheduler = SCHEDULES[self.schedule](
                optimizer, step_size=self.num_train_optimization_steps
            )  # PyTorch scheduler
        else:
            scheduler = SCHEDULES[self.schedule](
                optimizer,
                num_warmup_steps=self.warmup_steps,
                num_training_steps=self.num_train_optimization_steps,
            )  # transformers scheduler

        lr_scheduler_config = {"scheduler": scheduler, "interval": "step", "frequency": 1}
        return {"optimizer": optimizer, "lr_scheduler": lr_scheduler_config}


class SequenceTaggingDistillationBase(TaskSpecificDistillation):
    def __init__(
        self,
        temperature: float,
        warmup_steps: int,
        learning_rate: float,
        weight_decay: float,
        batch_size: int,
        accumulate_grad_batches: int,
        max_epochs: int,
        max_length: int,
        data_dir: str,
        label_list: Union[list, ListConfig],
        student_model_path: str,
        teacher_model_path: str,
        num_workers: int,
        schedule: Optional[str] = None,
        metric: str = "Default",
    ):
        """Base class for sequence tagging (task-specific) distillation steps.

        :param temperature:
        :param warmup_steps:
        :param learning_rate:
        :param weight_decay:
        :param batch_size:
        :param accumulate_grad_batches:
        :param max_epochs:
        :param data_dir:
        :param label_list:
        :param student_model_path:
        :param teacher_model_path:
        :param num_workers:
        :param schedule:
        :param metric:
        """
        self.processor = NerProcessor()
        self.data_dir = data_dir
        self.max_length = max_length
        self.tokenizer: Union[
            PreTrainedTokenizer, PreTrainedTokenizerFast
        ] = AutoTokenizer.from_pretrained(student_model_path)
        super().__init__(
            schedule=schedule,
            accumulate_grad_batches=accumulate_grad_batches,
            temperature=temperature,
            warmup_steps=warmup_steps,
            weight_decay=weight_decay,
            learning_rate=learning_rate,
            max_epochs=max_epochs,
            batch_size=batch_size,
        )

        self.num_workers = num_workers
        self.label_list: list[str]
        if isinstance(label_list, ListConfig):
            self.label_list = cast(list[str], OmegaConf.to_container(label_list))
        else:
            self.label_list = label_list
        self.num_labels = len(label_list)
        self.label_map = {label: i for i, label in enumerate(label_list)}
        self.metric = metric

        self.teacher_model = TinyBertForSequenceTagging.from_pretrained(
            teacher_model_path, num_labels=self.num_labels
        )
        self.student_model = TinyBertForSequenceTagging.from_pretrained(
            student_model_path, num_labels=self.num_labels
        )

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.train_dataloader
        </common/lightning_module.rst#train-dataloader>`\\ ."""
        dataset = NerDataset(
            tokenizer=self.tokenizer,
            examples=self.training_examples,
            label_map=self.label_map,
            max_length=self.max_length,
        )
        collator = DataCollatorForTokenClassification(tokenizer=self.tokenizer, padding=True)
        return DataLoader(
            dataset=dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            collate_fn=collator,
            pin_memory=True,
            persistent_workers=True,
        )

    def get_training_examples(self) -> list[InputExample]:
        return self.processor.get_train_examples(self.data_dir)

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.val_dataloader
        </common/lightning_module.rst#val-dataloader>`\\ ."""
        examples = self.processor.get_dev_examples(self.data_dir)
        dataset = NerDataset(
            tokenizer=self.tokenizer,
            examples=examples,
            label_map=self.label_map,
            max_length=self.max_length,
        )
        collator = DataCollatorForTokenClassification(tokenizer=self.tokenizer, padding=True)
        return DataLoader(
            dataset=dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            collate_fn=collator,
            pin_memory=True,
            persistent_workers=True,
        )


class SequenceTaggingDistillationForFinalLayer(SequenceTaggingDistillationBase):
    def __init__(
        self,
        temperature: float,
        warmup_steps: int,
        learning_rate: float,
        weight_decay: float,
        batch_size: int,
        accumulate_grad_batches: int,
        max_epochs: int,
        max_length: int,
        data_dir: str,
        label_list: Union[list, ListConfig],
        student_model_path: str,
        teacher_model_path: str,
        num_workers: int,
        schedule: Optional[str] = None,
        metric: str = "Default",
    ):
        """A class for sequence tagging (task-specific) final-layer
        distillation step.

        :param temperature:
        :param warmup_steps:
        :param learning_rate:
        :param weight_decay:
        :param batch_size:
        :param accumulate_grad_batches:
        :param max_epochs:
        :param data_dir:
        :param label_list:
        :param student_model_path:
        :param teacher_model_path:
        :param num_workers:
        :param schedule:
        :param metric:
        """
        super().__init__(
            temperature=temperature,
            warmup_steps=warmup_steps,
            weight_decay=weight_decay,
            learning_rate=learning_rate,
            batch_size=batch_size,
            accumulate_grad_batches=accumulate_grad_batches,
            max_epochs=max_epochs,
            max_length=max_length,
            data_dir=data_dir,
            label_list=label_list,
            student_model_path=student_model_path,
            teacher_model_path=teacher_model_path,
            num_workers=num_workers,
            schedule=schedule,
            metric=metric,
        )
        # Loss function: self.soft_cross_entropy for training, CrossEntropyLoss for validation
        self.loss = CrossEntropyLoss(ignore_index=IGNORE_IDX)
        self.save_hyperparameters()

    def soft_cross_entropy(self, predicts, targets):
        student_likelihood = torch.nn.functional.log_softmax(predicts, dim=-1)
        targets_prob = torch.nn.functional.softmax(targets, dim=-1)
        return (-targets_prob * student_likelihood).mean()

    def training_step(self, batch, batch_idx):
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.training_step
        </common/lightning_module.rst#training-step>`\\ ."""
        student_logits, student_atts, student_reps = self.student_model(
            input_ids=batch["input_ids"],
            token_type_ids=batch["token_type_ids"],
            attention_mask=batch["attention_mask"],
            is_student=True,
        )

        # self.teacher_model.eval()
        with torch.no_grad():
            teacher_logits, teacher_atts, teacher_reps = self.teacher_model(
                input_ids=batch["input_ids"],
                token_type_ids=batch["token_type_ids"],
                attention_mask=batch["attention_mask"],
            )

        loss = self.soft_cross_entropy(
            student_logits / self.temperature, teacher_logits / self.temperature
        )

        # Logging
        self.log("training_loss", loss, prog_bar=True, on_step=True)
        scheduler = self.lr_schedulers()
        assert isinstance(scheduler, LRScheduler)
        lr_list = scheduler.get_last_lr()
        self.log("lr", lr_list[0], prog_bar=True, on_step=True)
        if lr_list[0] != lr_list[1]:
            self.log("lr1", lr_list[1], prog_bar=True, on_step=True)

        return loss

    def validation_step(self, batch, batch_idx):
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.validation_step
        </common/lightning_module.rst#validation-step>`\\ ."""
        logits, _, _ = self.student_model(
            input_ids=batch["input_ids"],
            token_type_ids=batch["token_type_ids"],
            attention_mask=batch["attention_mask"],
            is_student=True,
        )

        loss = (
            self.loss(
                logits.view(-1, self.num_labels),
                batch["labels"].view(-1),
            )
            .mean()
            .item()
        )
        return {
            "loss": loss,
            "logits": logits.detach().cpu(),
            "label_ids": batch["labels"].detach().cpu(),
            "attention_mask": batch["attention_mask"].detach().cpu(),
        }

    def validation_epoch_end(self, val_step_outputs):
        """Implementation of :meth:`LightningModule.validation_epoch_end
        <pytorch_lightning.core.LightningModule.validation_epoch_end>`\\ ."""
        epoch_loss_mean = np.mean([x["loss"] for x in val_step_outputs])
        self.log(
            "validation_loss",
            epoch_loss_mean,
            prog_bar=True,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )

        preds, golds = [], []
        for output in val_step_outputs:

            attention_mask = output["attention_mask"]
            golds.extend(self.tensor_to_jagged_array(output["label_ids"], attention_mask))
            preds.extend(
                self.tensor_to_jagged_array(torch.argmax(output["logits"], dim=-1), attention_mask)
            )

        result = numeric_label_f1_score(preds=preds, golds=golds, label_list=self.label_list)
        self.log(
            self.metric, result, prog_bar=True, on_step=False, on_epoch=True, sync_dist=True
        )  # Micro F1

    def tensor_to_jagged_array(
        self, tensor: torch.Tensor, attention_mask: torch.Tensor
    ) -> list[list[int]]:
        result = []
        for arr, mask in zip(tensor.numpy(), attention_mask.numpy()):
            result.append(arr[0 : mask.sum()].tolist())
        return result


class SequenceTaggingDistillationForIntermediateLayer(SequenceTaggingDistillationBase):
    def __init__(
        self,
        temperature: float,
        warmup_steps: int,
        learning_rate: float,
        weight_decay: float,
        batch_size: int,
        accumulate_grad_batches: int,
        max_epochs: int,
        max_length: int,
        data_dir: str,
        label_list: Union[list, ListConfig],
        student_model_path: str,
        teacher_model_path: str,
        num_workers: int,
        schedule: Optional[str] = None,
        metric: str = "Default",
    ):
        """A class for sequence tagging (task-specific) intermediate-layer
        (Transformer, Embedding) distillation step.

        :param temperature:
        :param warmup_steps:
        :param learning_rate:
        :param weight_decay:
        :param batch_size:
        :param accumulate_grad_batches:
        :param max_epochs:
        :param data_dir:
        :param label_list:
        :param student_model_path:
        :param teacher_model_path:
        :param num_workers:
        :param schedule:
        :param metric:
        """
        super().__init__(
            temperature=temperature,
            warmup_steps=warmup_steps,
            weight_decay=weight_decay,
            learning_rate=learning_rate,
            batch_size=batch_size,
            accumulate_grad_batches=accumulate_grad_batches,
            max_epochs=max_epochs,
            max_length=max_length,
            data_dir=data_dir,
            label_list=label_list,
            student_model_path=student_model_path,
            teacher_model_path=teacher_model_path,
            num_workers=num_workers,
            schedule=schedule,
            metric=metric,
        )

        self.loss = MSELoss()
        self.save_hyperparameters()

    def _run_step(self, batch: Any) -> tuple[torch.Tensor, torch.Tensor]:
        """function for the training/validation step. Computes attention based
        distillation loss and hidden states based distillation loss.

        :param batch: The output of DataLoader.
        :return: A tuple of tensors. (rep_loss, att_loss)

            rep_loss: hidden states based distillation loss (includes embedding-layer distillation)
            att_loss: attention based distillation loss
        """
        student_logits, student_atts, student_reps = self.student_model(
            input_ids=batch["input_ids"],
            token_type_ids=batch["token_type_ids"],
            attention_mask=batch["attention_mask"],
            is_student=True,
        )

        # self.teacher_model.eval()
        with torch.no_grad():
            teacher_logits, teacher_atts, teacher_reps = self.teacher_model(
                input_ids=batch["input_ids"],
                token_type_ids=batch["token_type_ids"],
                attention_mask=batch["attention_mask"],
            )

        teacher_layer_num = len(teacher_atts)
        student_layer_num = len(student_atts)
        assert teacher_layer_num % student_layer_num == 0
        layers_per_block = int(teacher_layer_num / student_layer_num)

        att_loss = torch.Tensor([0.0])
        rep_loss = torch.Tensor([0.0])

        new_teacher_atts = [
            teacher_atts[i * layers_per_block + layers_per_block - 1]
            for i in range(student_layer_num)
        ]

        for student_att, teacher_att in zip(student_atts, new_teacher_atts):
            student_att_cond = torch.where(
                student_att <= -1e2, torch.zeros_like(student_att), student_att
            )
            teacher_att_cond = torch.where(
                teacher_att <= -1e2, torch.zeros_like(teacher_att), teacher_att
            )

            att_loss += self.loss(student_att_cond, teacher_att_cond)

        new_teacher_reps = [
            teacher_reps[i * layers_per_block] for i in range(student_layer_num + 1)
        ]

        for student_rep, teacher_rep in zip(student_reps, new_teacher_reps):
            rep_loss += self.loss(student_rep, teacher_rep)

        return rep_loss, att_loss

    def training_step(self, batch, batch_idx):
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.training_step
        </common/lightning_module.rst#training-step>`\\ ."""
        rep_loss, att_loss = self._run_step(batch)
        loss = rep_loss + att_loss

        # Logging
        self.log("training_loss", loss, on_step=True)
        self.log("att_loss", att_loss, on_step=True)
        self.log("rep_loss", rep_loss, on_step=True)
        scheduler = self.lr_schedulers()
        assert isinstance(scheduler, LRScheduler)
        lr_list = scheduler.get_last_lr()
        self.log("lr", lr_list[0], prog_bar=True, on_step=True)
        if lr_list[0] != lr_list[1]:
            self.log("lr1", lr_list[1], prog_bar=True, on_step=True)

        return loss

    def validation_step(self, batch, batch_idx):
        """Implementation of
        :external+pytorch_lightning:ref:`LightningModule.validation_step
        </common/lightning_module.rst#validation-step>`\\ ."""
        rep_loss, att_loss = self._run_step(batch)
        loss = rep_loss + att_loss
        return {
            "loss": loss.detach().cpu(),
            "rep_loss": rep_loss.detach().cpu(),
            "att_loss": att_loss.detach().cpu(),
        }

    def validation_epoch_end(self, val_step_outputs):
        """Implementation of :meth:`LightningModule.validation_epoch_end
        <pytorch_lightning.core.LightningModule.validation_epoch_end>`\\ ."""

        epoch_loss_mean = np.mean([x["loss"] for x in val_step_outputs])
        self.log(
            self.metric,
            epoch_loss_mean,
            prog_bar=True,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )

        epoch_rep_loss_mean = np.mean([x["rep_loss"] for x in val_step_outputs])
        self.log(
            "val_rep_loss",
            epoch_rep_loss_mean,
            prog_bar=True,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )
        epoch_att_loss_mean = np.mean([x["att_loss"] for x in val_step_outputs])
        self.log(
            "val_att_loss",
            epoch_att_loss_mean,
            prog_bar=True,
            on_step=False,
            on_epoch=True,
            sync_dist=True,
        )



import argparse
import json
import logging
import os
import shutil
import subprocess
from dataclasses import dataclass, field
from logging.config import fileConfig
from pathlib import Path
from typing import Optional

import ray
from hydra import initialize_config_dir, compose
from omegaconf import DictConfig

from kazu import __version__ as kazu_version


@dataclass
class BuildConfiguration:
    """Dataclass that controls how a base model pack and config should be
    merged with a target model pack."""

    #: should this model pack use the base config as a starting point?
    requires_base_config: bool
    #: what resource directories should this model pack include?
    #: structure is <parent_directory>:[paths within parent]
    resources: dict[str, list[str]]
    #: does this model pack have its own config dir? (if used with use_base_config
    #: these will override any config files from the base config)
    has_own_config: bool
    #: should acceptance tests be run?
    run_acceptance_tests: bool = False
    #: should consistency checks be run on the gold standard?
    run_consistency_checks: bool = False
    #: Whether resources (e.g. model binaries) are required to build this model pack
    #: This will be set automatically based on the values of the other fields,
    #: it's not available to set when instantiating the class.
    requires_resources: bool = field(init=False)

    def __post_init__(self):
        if len(self.resources) > 0:
            self.requires_resources = True
        else:
            self.requires_resources = False


class ModelPackBuildError(Exception):
    pass


class ModelPackBuilder:
    def __init__(
        self,
        logging_config_path: Optional[Path],
        target_model_pack_path: Path,
        kazu_version: str,
        build_dir: Path,
        maybe_base_configuration_path: Optional[Path],
        skip_tests: bool,
        zip_pack: bool,
    ):
        """A ModelPackBuilder is a helper class to assist in the building of a
        model pack.

        .. danger::
           WARNING! since this class will configure the kazu global cache, executing
           multiple builds within the same python process could potentially lead to the
           pollution of the cache. This is because the KAZU_MODEL_PACK env variable
           is modified by this object, which should normally not happen. Rather
           than instantiating this object directly, one should instead use
           :func:`.build_all_model_packs`\\, which will control this process for you.


        :param logging_config_path: passed to :func:`logging.config.fileConfig`
        :param target_model_pack_path: path to model pack to process
        :param kazu_version: version of kazu used to generate model pack
        :param build_dir: build the pack in this directory
        :param maybe_base_configuration_path: if this pack requires the base configuration, specify path
        :param skip_tests: don't run any tests
        :param zip_pack: zip the pack at the end (requires the 'zip' CLI tool)
        """
        if logging_config_path is not None:
            fileConfig(logging_config_path)
        self.logger = logging.getLogger(__name__)
        self.zip_pack = zip_pack
        self.skip_tests = skip_tests
        self.maybe_base_configuration_path = maybe_base_configuration_path
        self.build_dir = build_dir
        self.kazu_version = kazu_version
        self.target_model_pack_path = target_model_pack_path
        self.model_pack_build_path = self.build_dir.joinpath(self.target_model_pack_path.name)
        os.environ["KAZU_MODEL_PACK"] = str(self.model_pack_build_path)
        self.build_config = self.load_build_configuration()

    def __repr__(self):
        """For nice log messages."""
        return f"ModelPackBuilder({self.target_model_pack_path.name})"

    def build_model_pack(self) -> Path:
        """Execute the build process.

        :return: path of new pack
        """

        self.logger.info("building model pack at %s", self.model_pack_build_path)
        self.model_pack_build_path.mkdir()
        self.apply_merge_configurations()

        self.clear_cached_resources_from_model_pack_dir()
        # local import so the cache is correctly configured with KAZU_MODEL_PACK
        from kazu.utils.constants import HYDRA_VERSION_BASE
        from kazu.steps.other.cleanup import StripMappingURIsAction

        with initialize_config_dir(
            version_base=HYDRA_VERSION_BASE,
            config_dir=str(self.model_pack_build_path.joinpath("conf")),
        ):
            cfg = compose(
                config_name="config",
                overrides=["hydra/job_logging=none", "hydra/hydra_logging=none"],
            )
            ModelPackBuilder.build_caches(cfg)
            if not self.skip_tests:
                # our annotations expect URI stripping, so if cleanup actions
                # are specified, ensure this is configured.
                # Note, this may cause unexpected behaviour if CleanupActions
                # is configured with anything other than 'default'?
                if "CleanupActions" in cfg:
                    if StripMappingURIsAction.__name__ not in cfg.CleanupActions:
                        self.logger.warning(
                            "%s will be overridden for acceptance tests.",
                            StripMappingURIsAction.__name__,
                        )
                        cfg = compose(
                            config_name="config",
                            overrides=[
                                "hydra/job_logging=none",
                                "hydra/hydra_logging=none",
                                "CleanupActions=[default,uri_stripping]",
                            ],
                        )

                # local import so the cache is correctly configured with KAZU_MODEL_PACK
                from kazu.annotation.acceptance_test import (
                    execute_full_pipeline_acceptance_test,
                    check_annotation_consistency,
                )

                if self.build_config.run_consistency_checks:
                    check_annotation_consistency(cfg)
                if self.build_config.run_acceptance_tests:
                    execute_full_pipeline_acceptance_test(cfg)
            self.report_tested_dependencies()
            if self.zip_pack:
                self.zip_model_pack()

        return self.model_pack_build_path

    def load_build_configuration(self) -> BuildConfiguration:
        """Try to load a build configuration from the model pack root.

        The merge configuration should be a json file called build_config.json.

        :raises ModelPackBuildError: if the merge config isn't found at the
            expected path
        """
        config_path = self.target_model_pack_path.joinpath("build_config.json")
        if not config_path.exists():
            raise ModelPackBuildError(f"no merge config found at {str(config_path)}")

        with open(config_path, "r") as f:
            data = json.load(f)
        return BuildConfiguration(**data)

    def apply_merge_configurations(self):

        # copy the target pack to the target build dir
        shutil.copytree(
            str(self.target_model_pack_path), str(self.model_pack_build_path), dirs_exist_ok=True
        )
        # copy base config if required
        if self.build_config.requires_base_config:
            if self.maybe_base_configuration_path is None:
                raise ModelPackBuildError(
                    f"merge config asked for base configuration path but none was provided: {self.build_config}"
                )
            self.model_pack_build_path.joinpath("conf").mkdir(exist_ok=True)
            shutil.copytree(
                str(self.maybe_base_configuration_path),
                str(self.model_pack_build_path.joinpath("conf")),
                dirs_exist_ok=True,
            )
        if self.build_config.has_own_config:
            # and copy target config to override required elements (if required)
            shutil.copytree(
                str(self.target_model_pack_path.joinpath("conf")),
                str(self.model_pack_build_path.joinpath("conf")),
                dirs_exist_ok=True,
            )

        if self.build_config.requires_resources:
            self.copy_resources_to_target()

    def copy_resources_to_target(self):

        for parent_dir_str, resource_list in self.build_config.resources.items():
            parent_dir_path = Path(parent_dir_str)
            assert parent_dir_path.is_dir(), f"{parent_dir_str} is not a directory"
            for resource_path in resource_list:
                full_path = parent_dir_path.joinpath(resource_path).absolute()
                target_dir = self.model_pack_build_path.joinpath(resource_path)
                if full_path.is_dir():
                    shutil.copytree(full_path, target_dir, dirs_exist_ok=True)
                else:
                    shutil.copy(full_path, target_dir)

    def clear_cached_resources_from_model_pack_dir(self):
        """Delete any cached data from the input path.

        :return:
        """
        # local import so the cache is correctly configured with KAZU_MODEL_PACK
        from kazu.utils.caching import kazu_disk_cache

        kazu_disk_cache.clear()

        maybe_spacy_pipeline = self.model_pack_build_path.joinpath("spacy_pipeline")
        if maybe_spacy_pipeline.exists():
            shutil.rmtree(maybe_spacy_pipeline)

    def zip_model_pack(self):
        """Call the zip subprocess to compress model pack (requires zip on CLI)
        also moves it to parent dir.

        :return:
        """
        model_pack_name = f"{self.model_pack_build_path.name}-v{self.kazu_version}.zip"
        self.logger.info(f"zipping model pack {model_pack_name}")
        parent_directory = self.model_pack_build_path.parent
        model_pack_name_with_version = model_pack_name.removesuffix(".zip")
        subprocess.run(
            # make a symlink so the top-level directory in the resulting zip file
            # has the version of the model pack in it
            ["ln", "-s", self.model_pack_build_path.name, model_pack_name_with_version],
            cwd=parent_directory,
        )
        subprocess.run(
            ["zip", "-r", model_pack_name, model_pack_name_with_version], cwd=parent_directory
        )

    @staticmethod
    def build_caches(cfg: DictConfig) -> None:
        """Execute all processed required to build model pack caches.

        :param cfg:
        :return:
        """
        from kazu.pipeline import load_steps_and_log_memory_usage

        load_steps_and_log_memory_usage(cfg)

    def report_tested_dependencies(self):
        dependencies = subprocess.check_output("pip freeze", shell=True).decode("utf-8")
        with self.model_pack_build_path.joinpath("tested_dependencies.txt").open(mode="w") as f:
            f.write(dependencies)


@ray.remote(num_cpus=1)
class ModelPackBuilderActor(ModelPackBuilder):
    pass


def build_all_model_packs(
    maybe_base_configuration_path: Optional[Path],
    model_pack_paths: list[Path],
    zip_pack: bool,
    output_dir: Path,
    skip_tests: bool,
    logging_config_path: Optional[Path],
    max_parallel_build: Optional[int],
) -> None:
    """Build multiple model packs.

    :param maybe_base_configuration_path: Path to the base configuration, if required
    :param model_pack_paths: list of paths to model pack resources
    :param zip_pack: should the packs be zipped at the end?
    :param output_dir: directory to build model packs in
    :param skip_tests: don't run any tests
    :param logging_config_path: passed to logging.config.fileConfig
    :param max_parallel_build: build at most this many model packs simultaneously.
        If None, use all available CPUs
    :return:
    """
    if not output_dir.is_dir():
        raise ModelPackBuildError(f"{str(output_dir)} is not a directory")
    if len(list(output_dir.iterdir())) > 0:
        raise ModelPackBuildError(f"{str(output_dir)} is not empty")

    runtime_env = {"env_vars": {"PL_DISABLE_FORK": str(1), "TOKENIZERS_PARALLELISM": "false"}}
    ray.init(
        num_cpus=max_parallel_build,
        runtime_env=runtime_env,
        configure_logging=False,
        log_to_driver=True,
    )
    max_parallel_build = (
        max_parallel_build if max_parallel_build is not None else ray.cluster_resources()["CPU"]
    )
    futures: list[ray.ObjectRef] = []
    for model_pack_path in model_pack_paths:
        builder = ModelPackBuilderActor.remote(  # type: ignore[attr-defined]
            logging_config_path=logging_config_path,
            maybe_base_configuration_path=maybe_base_configuration_path,
            kazu_version=kazu_version,
            zip_pack=zip_pack,
            target_model_pack_path=model_pack_path,
            build_dir=output_dir,
            skip_tests=skip_tests,
        )
        futures.append(builder.build_model_pack.remote())
        while len(futures) >= max_parallel_build:
            futures = wait_for_model_pack_completion(futures)

    while len(futures) != 0:
        futures = wait_for_model_pack_completion(futures)


def wait_for_model_pack_completion(futures: list[ray.ObjectRef]) -> list[ray.ObjectRef]:
    # Returns the first ObjectRef that is ready.
    finished, futures = ray.wait(futures, num_returns=1, timeout=180.0 * 60.0)
    result = ray.get(finished[0])
    print(f"model pack {result} build complete")
    return futures


if __name__ == "__main__":
    description = """Build and test model packs. This script takes a list of model
pack directories, and creates model packs with a number of options. Depending on
how it is called, one or more of the following may be required:

1) a conf dir (Hydra configuration for the model pack), containing
 a) configurations for a pipeline to be created
 b) a LabelStudioManager configuration to run acceptance tests or consistency checks, pointing to a running
  and network-accessible instance of LabelStudio
2) an acceptance_criteria.json in the path <model_pack_root>/acceptance_criteria.json in order for the
  acceptance tests to run. This file should specify the thresholds per NER class/Ontology for NER/linking
  respectively (see the provided model pack for an example)
3) a merge config in the path <model_pack_root>/build_config.json, which determines which elements
  of the base configuration and model pack should be used

Note, if the environment variable KAZU_MODEL_PACK_BUILD_RESOURCES_PATH is set, this script will change it's
working directory to that location before doing anything. This will change the interpretation of relative
paths in a model packs build_config.json.
"""

    working_dir_str = os.getenv("KAZU_MODEL_PACK_BUILD_RESOURCES_PATH")
    if working_dir_str is not None:
        path = Path(working_dir_str)
        if path.is_dir():
            print(
                f"KAZU_MODEL_PACK_BUILD_RESOURCES_PATH is set to {working_dir_str}. Changing to"
                f" this path. This will affect relative paths in build_config.json"
            )
            os.chdir(path)
        else:
            raise ValueError(
                f"KAZU_MODEL_PACK_BUILD_RESOURCES_PATH is set to {working_dir_str}, but this is not a directory"
            )

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "--base_configuration_path",
        type=Path,
        required=False,
        help="""Path to the base configuration to use if a model pack requires it (generally
        this should be <kazu repo root>/kazu/conf)""",
    )
    parser.add_argument(
        "--model_packs_to_build",
        type=Path,
        nargs="+",
        required=True,
        help="paths to model packs to build",
    )
    parser.add_argument(
        "--zip_model_pack",
        action="store_true",
        help="should the model pack be zipped? (requires command line tool 'zip')",
    )
    parser.add_argument(
        "--model_pack_output_path",
        type=Path,
        required=True,
        help="create model packs at this location",
    )
    parser.add_argument(
        "--skip_tests",
        action="store_true",
        help="don't run any tests",
    )
    parser.add_argument(
        "--logging_config_path",
        type=Path,
        required=False,
        help="path to a logging config file, if required",
    )
    parser.add_argument(
        "--max_parallel_build",
        type=int,
        required=False,
        help="build at most this many model packs simultaneously. If None, use all available CPUs",
    )

    args = parser.parse_args()

    build_all_model_packs(
        maybe_base_configuration_path=args.base_configuration_path,
        model_pack_paths=args.model_packs_to_build,
        zip_pack=args.zip_model_pack,
        output_dir=args.model_pack_output_path,
        skip_tests=args.skip_tests,
        logging_config_path=args.logging_config_path,
        max_parallel_build=args.max_parallel_build,
    )
